/****************************************************************************
**  CUBE        http://www.score-p.org/                                    **
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2014                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <cstring>
#include <iostream>
#include <cassert>
#include <cmath>
// #ifndef CUBE_COMPRESSED
#include <fstream>
// #else
#include "CubeZfstream.h"
// #endif

#include <QMouseEvent>
#include <QString>
#include <QHeaderView>
#include <QInputDialog>
#include <QMessageBox>
#include <QFileDialog>
#include <QTextStream>
#include <QHBoxLayout>
#include <QPushButton>
#include <QApplication>
#include <QSettings>
#include <QScrollBar>
#include <QtNetwork>
#include <QDialogButtonBox>
#include <QPainter>
#include <QRadioButton>
#include <QFontDatabase>
#include <QRegExp>
#include <QTextEdit>
#include <QClipboard>
#include <QFontDialog>
#include <QFontMetrics>


#include "CallTreeWidget.h"
#include "TypedWidget.h"
#include "TabWidget.h"
#include "PrecisionWidget.h"
#include "ValueWidget.h"
#include "HelpBrowser.h"
#include "ScrollArea.h"
#include "MessageWidget.h"

#ifdef WITH_EXPERIMENTAL_VIEWS
#  include "2DPlotWidget.h"
#  include "ColorMapPlotWidget.h"
#endif   /* WITH_EXPERIMENTAL_VIEWS */

#include "Cube.h"
#include "CubeMetric.h"
#include "CubeCnode.h"
#include "CubeRegion.h"
#include "CubeMachine.h"
#include "CubeNode.h"
#include "CubeProcess.h"
#include "CubeThread.h"
#include "CubeVertex.h"
#include "CubeError.h"
#include "CubeValues.h"

using namespace std;
using namespace cube;

//special case:
//compute the current value for a METRIC tree widget item
//if the metric tree is on LEFT
//
void
CallTreeWidget::computeValue( list_of_metrics& mv, TreeWidgetItem* item )
{
    if ( !item->isInvalid() && !item->isExpanded() )
    {
        return;
    }

    if ( item->isExpanded() && item->getHiddenIterations().size() > 0 )
    {
        list_of_sysresources empty;
        computeValueHiddenIterationsFirstLevel( mv, empty, item );
        computeValueHiddenIterations( mv, empty, item, item->getHiddenIterations() );
        return;
    }

    if ( item->isInvalid() )
    {
        assert( item->type == CALLITEM );

        cube::Cnode* cnode = ( ( cube::Cnode* )( item->cubeObject ) );

        item->setValues( cube->get_sev_adv( mv, cnode, cube::CUBE_CALCULATE_INCLUSIVE ),
                         cube->get_sev_adv( mv, cnode, cube::CUBE_CALCULATE_EXCLUSIVE ) );
    }

    if ( item->isExpanded() )
    {
        for ( int i = 0; i < ( int )( item->childCount() ); i++ )
        {
            computeValue( mv, item->child( i ) );
        }
    }
}

//special case:
//compute the current value for calltreewidget item
//if the calltree is the last one
//
void
CallTreeWidget::computeValue( list_of_metrics& mv, list_of_sysresources& sv, TreeWidgetItem* item )
{
    if ( !item->isInvalid() && !item->isExpanded() )
    {
        return;
    }

    if ( item->isExpanded() && item->getHiddenIterations().size() > 0 )
    {
        computeValueHiddenIterations( mv, sv, item, item->getHiddenIterations() );
        return;
    }

    if ( item->isInvalid() )
    {
        assert( item->type == CALLITEM );

        cube::Cnode* cnode = ( ( cube::Cnode* )( item->cubeObject ) );

        item->setValues( cube->get_sev_adv( mv, sv, cnode, cube::CUBE_CALCULATE_INCLUSIVE ),
                         cube->get_sev_adv( mv, sv, cnode, cube::CUBE_CALCULATE_EXCLUSIVE ) );
    }

    if ( item->isExpanded() )
    {
        for ( int i = 0; i < ( int )( item->childCount() ); i++ )
        {
            computeValue( mv, sv, item->child( i ) );
        }
    }
}

/**
 * @brief CallTreeWidget::computeValueHiddenIterationsFirstLevel
 * If iterations are merged, the tree level with the iteration items is removed. The iterations items
 * may have exclusive values, which have to be added to the exclusive value of its parent.
 */
void
CallTreeWidget::computeValueHiddenIterationsFirstLevel( list_of_metrics& mv, list_of_sysresources& sv, TreeWidgetItem* item )
{
    cube::Cnode*   cnode = ( cube::Cnode* )( item->cubeObject );
    list_of_cnodes cnodesExcl;

    // add root item to cnode list
    cnode_pair cp;
    cp.first  = ( cube::Cnode* )( item->cubeObject );
    cp.second = CUBE_CALCULATE_EXCLUSIVE;
    cnodesExcl.push_back( cp );

    // add child items (iterations) to cnode list
    foreach( TreeWidgetItem * iteration, item->getHiddenIterations() )
    {
        cp.first = ( cube::Cnode* )( iteration->cubeObject );
        cnodesExcl.push_back( cp );
    }

    if ( sv.size() == 0 ) // METRIC tree is on left
    {
        cube::Value* total = cube->get_sev_adv( mv, cnode, cube::CUBE_CALCULATE_INCLUSIVE );
        cube::Value* own   = cube->get_sev_adv( mv, cnodesExcl );
        item->setValues( total, own );
    }
    else // CALL tree is last one
    {
        cube::Value* total = cube->get_sev_adv( mv, sv, cnode, cube::CUBE_CALCULATE_INCLUSIVE );
        cube::Value* own   = cube->get_sev_adv( mv, cnodesExcl, sv );
        item->setValues( total, own );
    }
}

/**
 * @brief CallTreeWidget::computeValueHiddenIterations calculates values for all children of a subtree with hidden iterations.
 * For each item to calculate, the corresponding items of the hidden iterations with the same name on the
 * same tree level are searched and the values are merged.
 * @param item TreeWidgetItem whose children will contain the merged values of all corresponding iterations
 * @param iterations vector with the TreeWidgetItems for each iteration
 */
void
CallTreeWidget::computeValueHiddenIterations( list_of_metrics& mv, list_of_sysresources& sv, TreeWidgetItem* item, const vector<TreeWidgetItem*>& iterations )
{
    // calculate values for all visible elements in "hide iterations" mode
    foreach( TreeWidgetItem * compactChild, item->getChildren() )
    {
        list_of_cnodes cnodesIncl;
        list_of_cnodes cnodesExcl;

        vector<TreeWidgetItem*> iterationChilden;

        // loop iterations, each iteration contains children with the same name as compactChild
        foreach( TreeWidgetItem * iteration, iterations )
        {
            for ( int j = 0; j < ( int )( iteration->childCount() ); j++ )
            {
                TreeWidgetItem* child = iteration->child( j );
                if ( child->name == compactChild->name ) // found same item in different iteration
                {
                    cnode_pair cp;
                    cp.first = ( cube::Cnode* )( child->cubeObject );

                    cp.second = cube::CUBE_CALCULATE_INCLUSIVE;
                    cnodesIncl.push_back( cp );
                    cp.second = CUBE_CALCULATE_EXCLUSIVE;
                    cnodesExcl.push_back( cp );

                    if ( compactChild->isExpanded() )
                    {
                        iterationChilden.push_back( child );
                    }
                    break;
                }
            }
        }
        if ( sv.size() == 0 ) // METRIC tree is on left
        {
            cube::Value* total = cube->get_sev_adv( mv, cnodesIncl );
            cube::Value* own   = cube->get_sev_adv( mv, cnodesExcl );
            compactChild->setValues( total, own );
        }
        else // CALL tree is last one
        {
            cube::Value* total = cube->get_sev_adv( mv, cnodesIncl, sv );
            cube::Value* own   = cube->get_sev_adv( mv, cnodesExcl, sv );
            compactChild->setValues( total, own );
        }

        if ( iterationChilden.size() > 0 )
        {
            computeValueHiddenIterations( mv, sv, compactChild, iterationChilden );
        }
    }
}

/**
 * @brief getSelectedFromHiddenIterations returns a list of items of the hidden iterations which correspond
 * to the selected merged item
 * @param iterations a vector of hidden iterations, starting with tree level after root
 * @param selected path to the selected item, contains the names of each ancestor of the selected item,
 * starting with the level after root
 * @return vector with the selected item for each iteration where an item with the same name at the same tree
 * level is found
 */
std::vector<TreeWidgetItem*>
getSelectedFromHiddenIterations( const std::vector<TreeWidgetItem*> iterations, QStringList& selected )
{
    std::vector<TreeWidgetItem*> nextLevel;
    std::vector<TreeWidgetItem*> ret;

    QString selectedName = selected.takeFirst();
    foreach( TreeWidgetItem * iter, iterations )
    {
        foreach( TreeWidgetItem * child, iter->getChildren() )
        {
            if ( child->name == selectedName ) // found selected item in different iteration
            {
                nextLevel.push_back( child );
            }
        }
    }
    if ( !selected.empty() )
    {
        ret = getSelectedFromHiddenIterations( nextLevel, selected );
    }
    else
    {
        ret = nextLevel;
    }
    return ret;
}

/**
 * @brief CallTreeWidget::selectedItemsAndIterations returns a list of all selected items like
 * myTreeWidget::selectedItems, if not hidden iteration is selected. If a hidden iteration is
 * selected, the corresponding item of all iterations is added to the list.
 */
std::vector<TreeWidgetItem*>
CallTreeWidget::selectedItemsAndIterations()
{
    return selectedAndIterations;
}


/**
 * @brief CallTreeWidget::onSelection fills selectedAndIterations with selectedItems and then
 * calls base class method
 * If iterations are hidden, all hidden siblings of the selected item are inserted
 */
void
CallTreeWidget::onSelection()
{
    selectedAndIterations.clear();
    foreach( TreeWidgetItem * sel, selected )
    {
        // search for parent item, which has hidden iterations
        // add the path to the child item to slist
        QStringList     slist;
        TreeWidgetItem* iterations = sel;
        while ( ( iterations->getHiddenIterations().size() == 0 ) && ( iterations->parent() != 0 ) )
        {
            slist.insert( 0, iterations->name );
            iterations = iterations->parent();
        }

        if ( iterations->getHiddenIterations().size() == 0 ) // no hidden iterations
        {
            selectedAndIterations.push_back( sel );
        }
        else if ( iterations == sel ) // special case: root of hidden iteration selected
        {
            selectedAndIterations.push_back( sel );
            if ( sel->isExpanded() ) // add exclusive values of the invisible level to exclusive root value
            {
                foreach( TreeWidgetItem * iteration, sel->getHiddenIterations() )
                {
                    iteration->setExpanded( true );
                    selectedAndIterations.push_back( iteration );
                }
            }
        }
        else // a child of hidden iterations has been selected => select child in all iterations
        {
            const std::vector<TreeWidgetItem*>& hidden = iterations->getHiddenIterations();

            // search the selected item in all iterations
            std::vector<TreeWidgetItem*> selHidden = getSelectedFromHiddenIterations( hidden, slist );
            foreach( TreeWidgetItem * isel, selHidden )
            {
                isel->setExpanded( sel->isExpanded() ); // for calculation: set expanded state of all iteration children
                selectedAndIterations.push_back( isel );
            }
        }
    }

    TreeWidget::onSelection();
}


//compute all absolute values for this tree's items
//
void
CallTreeWidget::computeBasicValues()
{
//         cout << "calltree SDFFSDG " << endl;
    //set all values of all items to 0.0
    initializeValues();

    int index, metricIndex, callIndex, systemIndex;
    getIndices( index, metricIndex, callIndex, systemIndex );


    if ( callIndex == 0 )
    {
        return;                 // situation, Calltree, XXX, XXX -> Calltree doesn't show anything
    }
    if ( callIndex == 1 && metricIndex == 2 )
    {
        return;                                        // situation, System, Calltree, Metric -> calltree doesnt show anything
    }
    if ( tabWidget->getMetricTree()->Items().size() == 0 ||
         tabWidget->getCallTree()->Items().size() == 0 ||
         tabWidget->getSystemTree()->Items().size() == 0
         )
    {
        return;
    }
    if ( tabWidget->getMetricTree()->selectedItems().size() == 0 ||
         tabWidget->getCallTree()->selectedItems().size() == 0 ||
         tabWidget->getSystemTree()->selectedItems().size() == 0
         )
    {
        return;
    }

    if ( callIndex == 1 )   // situation  Metric, Calltree, System -> standard
    {
        list_of_metrics              mv;
        std::vector<TreeWidgetItem*> sel_metrics = tabWidget->getMetricTree()->selectedItems();
        for ( std::vector<TreeWidgetItem*>::iterator miter = sel_metrics.begin(); miter != sel_metrics.end(); miter++ )
        {
            metric_pair mp;
            mp.first  = ( cube::Metric* )( ( *miter )->cubeObject );
            mp.second = ( ( *miter )->isExpanded() ) ? CUBE_CALCULATE_EXCLUSIVE : CUBE_CALCULATE_INCLUSIVE;
            if ( ( *miter )->childCount() == 0 )
            {
                mp.second = CUBE_CALCULATE_INCLUSIVE;
            }
            mv.push_back( mp );
        }
        for ( int i = 0; i < ( int )top.size(); i++ )
        {
            computeValue( mv, top[ i ] );
        }
    }

    if ( callIndex == 2 )  // situation System, Metric , Calltree or situation Metric, System, Calltree
    {
        list_of_metrics      mv;
        list_of_sysresources sv;

        std::vector<TreeWidgetItem*> sel_metrics = tabWidget->getMetricTree()->selectedItems();
        for ( std::vector<TreeWidgetItem*>::iterator miter = sel_metrics.begin(); miter != sel_metrics.end(); miter++ )
        {
            metric_pair mp;
            mp.first  = ( cube::Metric* )( ( *miter )->cubeObject );
            mp.second = ( ( *miter )->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
            if ( ( *miter )->childCount() == 0 )
            {
                mp.second = CUBE_CALCULATE_INCLUSIVE;
            }
            mv.push_back( mp );
        }
        std::vector<TreeWidgetItem*> sel_systemnodes = tabWidget->getSystemTree()->selectedItems();
        for ( std::vector<TreeWidgetItem*>::iterator siter = sel_systemnodes.begin(); siter != sel_systemnodes.end(); siter++ )
        {
            sysres_pair sp;
            sp.first  = ( cube::Sysres* )( ( *siter )->cubeObject );
            sp.second = ( ( *siter )->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
            sv.push_back( sp );
        }

        for ( int i = 0; i < ( int )top.size(); i++ )
        {
            computeValue( mv, sv, top[ i ] );
        }
    }

    for ( int i = 0; i < ( int )top.size(); i++ )
    {
        validateValue( top[ i ] );
    }
}

QIcon
CallTreeWidget::icon() const
{
    return QIcon( ":images/tree-icon.png" );
}
//end of computeBasicValues()
