/****************************************************************************
**  CUBE        http://www.score-p.org/                                    **
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2014                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/

#include <cstring>
#include <iostream>
#include <cassert>
#include <cmath>
// #ifndef CUBE_COMPRESSED
#include <fstream>
// #else
#include "CubeZfstream.h"
// #endif

#include <QMouseEvent>
#include <QString>
#include <QHeaderView>
#include <QInputDialog>
#include <QMessageBox>
#include <QFileDialog>
#include <QTextStream>
#include <QHBoxLayout>
#include <QPushButton>
#include <QApplication>
#include <QSettings>
#include <QScrollBar>
#include <QtNetwork>
#include <QDialogButtonBox>
#include <QPainter>
#include <QRadioButton>
#include <QFontDatabase>
#include <QRegExp>
#include <QTextEdit>
#include <QClipboard>
#include <QFontDialog>
#include <QFontMetrics>


#include "FlatTreeWidget.h"
#include "TypedWidget.h"
#include "TabWidget.h"
#include "PrecisionWidget.h"
#include "ValueWidget.h"
#include "HelpBrowser.h"
#include "ScrollArea.h"
#include "MessageWidget.h"

#include "Cube.h"
#include "CubeMetric.h"
#include "CubeCnode.h"
#include "CubeRegion.h"
#include "CubeMachine.h"
#include "CubeNode.h"
#include "CubeProcess.h"
#include "CubeThread.h"
#include "CubeVertex.h"
#include "CubeError.h"
#include "CubeValues.h"

using namespace std;
using namespace cube;



//special case:
//compute the current value for a METRIC tree widget item
//if the metric tree is on LEFT
//
void
FlatTreeWidget::computeValue( list_of_metrics& mv, TreeWidgetItem* item )
{
    if ( !item->isInvalid() && !item->isExpanded() )
    {
        return;
    }
    if ( item->isInvalid() )
    {
        assert( item->type == REGIONITEM );

        cube::Region* region = ( cube::Region* )( item->cubeObject );
        if ( region == NULL )
        {
            region = ( cube::Region* )( item->parent()->cubeObject );
            region->set_as_subroutines();
        }

        if ( item->totalValue_adv != NULL )
        {
            item->totalValue_adv->Free();
        }
        if ( item->ownValue_adv != NULL )
        {
            item->ownValue_adv->Free();
        }
        item->ownValue_adv   = cube->get_sev_adv( mv, region, cube::CUBE_CALCULATE_EXCLUSIVE );
        item->totalValue_adv = cube->get_sev_adv( mv, region, cube::CUBE_CALCULATE_INCLUSIVE );
        region->unset_as_subroutines();
        item->totalValue     =  ( item->totalValue_adv != NULL )  ? item->totalValue_adv->getDouble() : 0.;
        item->ownValue       =    ( item->ownValue_adv != NULL )    ? item->ownValue_adv->getDouble() : 0.;
        item->valueCollapsed = item->totalValue;
        item->valueExpanded  = item->ownValue;

        item->singleValue = ( item->ownValue_adv != NULL )    ? item->ownValue_adv->singleValue() : true;
    }
    if ( item->parent() != NULL )
    {
        item->rootValue = item->parent()->rootValue;
    }
    else
    {
        item->rootValue = item->totalValue;
    }
    if ( item->isExpanded() )
    {
        for ( int i = 0; i < ( int )( item->childCount() ); i++ )
        {
            computeValue( mv, item->child( i ) );
        }
    }
}


//special case:
//compute the current value for calltreewidget item
//if the calltree is the last one
//
void
FlatTreeWidget::computeValue( list_of_metrics& mv, list_of_sysresources& sv, TreeWidgetItem* item )
{
    if ( !item->isInvalid() && !item->isExpanded() )
    {
        return;
    }
    if ( item->isInvalid() )
    {
        assert( item->type == REGIONITEM );

        cube::Region* region = ( cube::Region* )( item->cubeObject );
        if ( region == NULL )
        {
            region = ( cube::Region* )( item->parent()->cubeObject );
            region->set_as_subroutines();
        }

        if ( item->totalValue_adv != NULL )
        {
            item->totalValue_adv->Free();
        }
        if ( item->ownValue_adv != NULL )
        {
            item->ownValue_adv->Free();
        }
        item->ownValue_adv   = cube->get_sev_adv( mv, sv, region, cube::CUBE_CALCULATE_EXCLUSIVE );
        item->totalValue_adv = cube->get_sev_adv( mv, sv, region, cube::CUBE_CALCULATE_INCLUSIVE );
        region->unset_as_subroutines();
        item->totalValue     = ( item->totalValue_adv != NULL ) ? item->totalValue_adv->getDouble() : 0.;
        item->ownValue       = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->getDouble() : 0.;
        item->valueCollapsed = item->totalValue;
        item->valueExpanded  = item->ownValue;

        item->singleValue = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->singleValue() : true;
    }
    if ( item->parent() != NULL )
    {
        item->rootValue = item->parent()->rootValue;
    }
    else
    {
        item->rootValue = item->totalValue;
    }
    if ( item->isExpanded() )
    {
        for ( int i = 0; i < ( int )( item->childCount() ); i++ )
        {
            computeValue( mv, sv, item->child( i ) );
        }
    }
}














//compute all absolute values for this tree's items
//
void
FlatTreeWidget::computeBasicValues()
{
//             cout << "flat SDFFSDG " << endl;
    //set all values of all items to 0.0
    initializeValues();


    int index, metricIndex, callIndex, systemIndex;
    getIndices( index, metricIndex, callIndex, systemIndex );




    if ( callIndex == 0 )
    {
        return;                 // situation, Calltree, XXX, XXX -> Calltree doesn't show anything
    }
    if ( callIndex == 1 && metricIndex == 2 )
    {
        return;                                        // situation, System, Calltree, Metric -> calltree doesnt show anything
    }
    if ( tabWidget->getMetricTree()->Items().size() == 0 ||
         tabWidget->getCallTree()->Items().size() == 0 ||
         tabWidget->getSystemTree()->Items().size() == 0
         )
    {
        return;
    }
    if ( tabWidget->getMetricTree()->selectedItems().size() == 0 ||
         tabWidget->getCallTree()->selectedItems().size() == 0 ||
         tabWidget->getSystemTree()->selectedItems().size() == 0
         )
    {
        return;
    }

    if ( callIndex == 1 ) // situation  Metric, Calltree, System -> standard
    {
        list_of_metrics mv;
        mv.clear();
        std::vector<TreeWidgetItem*> sel_metrics = tabWidget->getMetricTree()->selectedItems();
        for ( std::vector<TreeWidgetItem*>::iterator miter = sel_metrics.begin(); miter != sel_metrics.end(); miter++ )
        {
            metric_pair mp;
            mp.first  = ( cube::Metric* )( ( *miter )->cubeObject );
            mp.second = ( ( *miter )->isExpanded() ) ? CUBE_CALCULATE_EXCLUSIVE : CUBE_CALCULATE_INCLUSIVE;
            mv.push_back( mp );
        }

        for ( int i = 0; i < ( int )top.size(); i++ )
        {
            computeValue( mv, top[ i ] );
        }
    }
    if ( callIndex == 2 )  // situation System, Metric , Flattree or situation Metric, System, Flattree
    {
        list_of_metrics      mv;
        list_of_sysresources sv;

        std::vector<TreeWidgetItem*> sel_metrics = tabWidget->getMetricTree()->selectedItems();
        for ( std::vector<TreeWidgetItem*>::iterator miter = sel_metrics.begin(); miter != sel_metrics.end(); miter++ )
        {
            metric_pair mp;
            mp.first  = ( cube::Metric* )( ( *miter )->cubeObject );
            mp.second = ( ( *miter )->isExpanded() ) ? CUBE_CALCULATE_EXCLUSIVE : CUBE_CALCULATE_INCLUSIVE;
            mv.push_back( mp );
        }
        std::vector<TreeWidgetItem*> sel_systemnodes = tabWidget->getSystemTree()->selectedItems();
        for ( std::vector<TreeWidgetItem*>::iterator siter = sel_systemnodes.begin(); siter != sel_systemnodes.end(); siter++ )
        {
            sysres_pair sp;
            sp.first  = ( cube::Sysres* )( ( *siter )->cubeObject );
            sp.second = ( ( *siter )->isExpanded() ) ? CUBE_CALCULATE_EXCLUSIVE : CUBE_CALCULATE_INCLUSIVE;
            sv.push_back( sp );
        }
        for ( int i = 0; i < ( int )top.size(); i++ )
        {
            computeValue( mv, sv, top[ i ] );
        }
//
    }




    for ( int i = 0; i < ( int )top.size(); i++ )
    {
        validateValue( top[ i ] );
    }
}

QIcon
FlatTreeWidget::icon() const
{
    return QIcon( ":images/list-icon.png" );
}
//end of computeBasicValues()
