/****************************************************************************
**  CUBE        http://www.score-p.org/                                    **
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2014                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/

#include <cstring>
#include <iostream>
#include <cassert>
#include <cmath>
// #ifndef CUBE_COMPRESSED
#include <fstream>
// #else
#include "CubeZfstream.h"
// #endif

#include <QMouseEvent>
#include <QString>
#include <QHeaderView>
#include <QInputDialog>
#include <QMessageBox>
#include <QFileDialog>
#include <QTextStream>
#include <QHBoxLayout>
#include <QPushButton>
#include <QApplication>
#include <QSettings>
#include <QScrollBar>
#include <QtNetwork>
#include <QDialogButtonBox>
#include <QPainter>
#include <QRadioButton>
#include <QFontDatabase>
#include <QRegExp>
#include <QTextEdit>
#include <QClipboard>
#include <QFontDialog>
#include <QFontMetrics>


#include "MetricTreeWidget.h"
#include "CallTreeWidget.h"
#include "TypedWidget.h"
#include "TabWidget.h"
#include "PrecisionWidget.h"
#include "ValueWidget.h"
#include "HelpBrowser.h"
#include "ScrollArea.h"
#include "MessageWidget.h"

#ifdef WITH_EXPERIMENTAL_VIEWS
#  include "2DPlotWidget.h"
#  include "ColorMapPlotWidget.h"
#endif   /* WITH_EXPERIMENTAL_VIEWS */

#include "Cube.h"
#include "CubeMetric.h"
#include "CubeCnode.h"
#include "CubeRegion.h"
#include "CubeMachine.h"
#include "CubeNode.h"
#include "CubeProcess.h"
#include "CubeThread.h"
#include "CubeVertex.h"
#include "CubeError.h"
#include "CubeValues.h"

using namespace std;
using namespace cube;





//special case:
//compute the current value for a METRIC tree widget item
//if the metric tree is on LEFT
//
void
MetricTreeWidget::computeValue( TreeWidgetItem* item )
{
    if ( !item->isInvalid() && !item->isExpanded() )
    {
        return;
    }
    if ( item->isInvalid() )
    {
        assert( item->type == METRICITEM );

        cube::Metric* metric = ( ( cube::Metric* )( item->cubeObject ) );

        if ( item->totalValue_adv != NULL )
        {
            item->totalValue_adv->Free();
        }
        if ( item->ownValue_adv != NULL )
        {
            item->ownValue_adv->Free();
        }

        item->totalValue_adv = cube->get_sev_adv( metric, cube::CUBE_CALCULATE_INCLUSIVE );
        item->ownValue_adv   = cube->get_sev_adv( metric, cube::CUBE_CALCULATE_EXCLUSIVE );
        item->totalValue     = ( item->totalValue_adv != NULL ) ? item->totalValue_adv->getDouble() : 0.;
        item->ownValue       = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->getDouble() : 0.;
        item->valueCollapsed = item->totalValue;
        item->valueExpanded  = item->ownValue;
        item->singleValue    = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->singleValue() : true;
    }

    if ( item->parent() != NULL )
    {
        item->rootValue = item->parent()->rootValue;
    }
    else
    {
        item->rootValue = item->totalValue;
    }


    if ( item->isExpanded() )
    {
        for ( int i = 0; i < ( int )( item->childCount() ); i++ )
        {
            computeValue( item->child( i ) );
        }
    }
}


//special case:
//compute the current value for a METRIC tree widget item
//if the metric tree is on center and calltree is on left
//
void
MetricTreeWidget::computeValue( list_of_cnodes& cv,  TreeWidgetItem* item )
{
    if ( !item->isInvalid() && !item->isExpanded() )
    {
        return;
    }

    if ( item->isInvalid() )
    {
        assert( item->type == METRICITEM );

        cube::Metric* metric = ( ( cube::Metric* )( item->cubeObject ) );

        if ( item->totalValue_adv != NULL )
        {
            item->totalValue_adv->Free();
        }
        if ( item->ownValue_adv != NULL )
        {
            item->ownValue_adv->Free();
        }
        item->totalValue_adv = cube->get_sev_adv( cv, metric, cube::CUBE_CALCULATE_INCLUSIVE );
        item->ownValue_adv   = cube->get_sev_adv( cv, metric, cube::CUBE_CALCULATE_EXCLUSIVE );
        item->totalValue     = ( item->totalValue_adv != NULL ) ? item->totalValue_adv->getDouble() : 0.;
        item->ownValue       = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->getDouble() : 0.;
        item->valueCollapsed = item->totalValue;
        item->valueExpanded  = item->ownValue;

        item->singleValue = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->singleValue() : true;
    }

    if ( item->parent() != NULL )
    {
        item->rootValue = item->parent()->rootValue;
    }
    else
    {
        item->rootValue = item->totalValue;
    }


    if ( item->isExpanded() )
    {
        for ( int i = 0; i < ( int )( item->childCount() ); i++ )
        {
            computeValue( cv, item->child( i ) );
        }
    }
}




//special case:
//compute the current value for a METRIC tree widget item
//if the metric tree is on center and calltree is on left
//
void
MetricTreeWidget::computeValue( list_of_regions& rv,  TreeWidgetItem* item )
{
    if ( !item->isInvalid() && !item->isExpanded() )
    {
        return;
    }

    if ( item->isInvalid() )
    {
        assert( item->type == METRICITEM );

        cube::Metric* metric = ( ( cube::Metric* )( item->cubeObject ) );

//      cube::CalculationFlavour cf = ((cnode_item)->isExpanded())?cube::CUBE_CALCULATE_EXCLUSIVE: cube::CUBE_CALCULATE_INCLUSIVE;
//      cube::Cnode * cnode = (cube::Cnode*)((cnode_item)->cubeObject);
        if ( item->totalValue_adv != NULL )
        {
            item->totalValue_adv->Free();
        }
        if ( item->ownValue_adv != NULL )
        {
            item->ownValue_adv->Free();
        }
        item->totalValue_adv = cube->get_sev_adv( rv, metric, cube::CUBE_CALCULATE_INCLUSIVE );
        item->ownValue_adv   = cube->get_sev_adv( rv, metric, cube::CUBE_CALCULATE_EXCLUSIVE );
        item->totalValue     = ( item->totalValue_adv != NULL ) ? item->totalValue_adv->getDouble() : 0.;
        item->ownValue       = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->getDouble() : 0.;
        item->valueCollapsed = item->totalValue;
        item->valueExpanded  = item->ownValue;

        item->singleValue = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->singleValue() : true;
    }

    if ( item->parent() != NULL )
    {
        item->rootValue = item->parent()->rootValue;
    }
    else
    {
        item->rootValue = item->totalValue;
    }


    if ( item->isExpanded() )
    {
        for ( int i = 0; i < ( int )( item->childCount() ); i++ )
        {
            computeValue( rv, item->child( i ) );
        }
    }
}




//special case:
//compute the current value for a METRIC tree widget item
//if the metric tree is on center and calltree is on left
//
void
MetricTreeWidget::computeValue( list_of_sysresources& sv,  TreeWidgetItem* item )
{
    if ( !item->isInvalid() && !item->isExpanded() )
    {
        return;
    }

    if ( item->isInvalid() )
    {
        assert( item->type == METRICITEM );

        cube::Metric* metric = ( ( cube::Metric* )( item->cubeObject ) );

//      cube::CalculationFlavour cf = ((cnode_item)->isExpanded())?cube::CUBE_CALCULATE_EXCLUSIVE: cube::CUBE_CALCULATE_INCLUSIVE;
//      cube::Cnode * cnode = (cube::Cnode*)((cnode_item)->cubeObject);
        if ( item->totalValue_adv != NULL )
        {
            item->totalValue_adv->Free();
        }
        if ( item->ownValue_adv != NULL )
        {
            item->ownValue_adv->Free();
        }
        item->totalValue_adv = cube->get_sev_adv( sv, metric, cube::CUBE_CALCULATE_INCLUSIVE );
        item->ownValue_adv   = cube->get_sev_adv( sv, metric, cube::CUBE_CALCULATE_EXCLUSIVE );
        item->totalValue     = ( item->totalValue_adv != NULL ) ? item->totalValue_adv->getDouble() : 0.;
        item->ownValue       = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->getDouble() : 0.;
        item->valueCollapsed = item->totalValue;
        item->valueExpanded  = item->ownValue;

        item->singleValue = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->singleValue() : true;
    }

    if ( item->parent() != NULL )
    {
        item->rootValue = item->parent()->rootValue;
    }
    else
    {
        item->rootValue = item->totalValue;
    }


    if ( item->isExpanded() )
    {
        for ( int i = 0; i < ( int )( item->childCount() ); i++ )
        {
            computeValue( sv, item->child( i ) );
        }
    }
}






//special case:
//compute the current value for a METRIC tree widget item
//if the metric tree is on center and calltree is on left
//
void
MetricTreeWidget::computeValue( list_of_regions& rv, list_of_sysresources& sv,  TreeWidgetItem* item )
{
    if ( !item->isInvalid() && !item->isExpanded() )
    {
        return;
    }

    if ( item->isInvalid() )
    {
        assert( item->type == METRICITEM );

        cube::Metric* metric = ( ( cube::Metric* )( item->cubeObject ) );

//      cube::CalculationFlavour cf = ((cnode_item)->isExpanded())?cube::CUBE_CALCULATE_EXCLUSIVE: cube::CUBE_CALCULATE_INCLUSIVE;
//      cube::Cnode * cnode = (cube::Cnode*)((cnode_item)->cubeObject);
        if ( item->totalValue_adv != NULL )
        {
            item->totalValue_adv->Free();
        }
        if ( item->ownValue_adv != NULL )
        {
            item->ownValue_adv->Free();
        }
        item->totalValue_adv = cube->get_sev_adv( rv, sv, metric, cube::CUBE_CALCULATE_INCLUSIVE );
        item->ownValue_adv   = cube->get_sev_adv( rv, sv, metric, cube::CUBE_CALCULATE_EXCLUSIVE );
        item->totalValue     = ( item->totalValue_adv != NULL ) ? item->totalValue_adv->getDouble() : 0.;
        item->ownValue       = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->getDouble() : 0.;
        item->valueCollapsed = item->totalValue;
        item->valueExpanded  = item->ownValue;

        item->singleValue = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->singleValue() : true;
    }

    if ( item->parent() != NULL )
    {
        item->rootValue = item->parent()->rootValue;
    }
    else
    {
        item->rootValue = item->totalValue;
    }


    if ( item->isExpanded() )
    {
        for ( int i = 0; i < ( int )( item->childCount() ); i++ )
        {
            computeValue( sv, item->child( i ) );
        }
    }
}








//special case:
//compute the current value for a METRIC tree widget item
//if the metric tree is on center and calltree is on left
//
void
MetricTreeWidget::computeValue( list_of_cnodes& cv, list_of_sysresources& sv,  TreeWidgetItem* item )
{
    if ( !item->isInvalid() && !item->isExpanded() )
    {
        return;
    }

    if ( item->isInvalid() )
    {
        assert( item->type == METRICITEM );

        cube::Metric* metric = ( ( cube::Metric* )( item->cubeObject ) );
        if ( item->totalValue_adv != NULL )
        {
            item->totalValue_adv->Free();
        }
        if ( item->ownValue_adv != NULL )
        {
            item->ownValue_adv->Free();
        }
        item->totalValue_adv = cube->get_sev_adv( cv, sv, metric, cube::CUBE_CALCULATE_INCLUSIVE );
        item->ownValue_adv   = cube->get_sev_adv( cv, sv, metric, cube::CUBE_CALCULATE_EXCLUSIVE );
        item->totalValue     = ( item->totalValue_adv != NULL ) ? item->totalValue_adv->getDouble() : 0.;
        item->ownValue       = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->getDouble() : 0.;
        item->valueCollapsed = item->totalValue;
        item->valueExpanded  = item->ownValue;

        item->singleValue = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->singleValue() : true;
    }

    if ( item->parent() != NULL )
    {
        item->rootValue = item->parent()->rootValue;
    }
    else
    {
        item->rootValue = item->totalValue;
    }


    if ( item->isExpanded() )
    {
        for ( int i = 0; i < ( int )( item->childCount() ); i++ )
        {
            computeValue( cv, sv, item->child( i ) );
        }
    }
}







//compute all absolute values for this tree's items
//
void
MetricTreeWidget::computeBasicValues()
{
//     cout << " SDFFSDG " << endl;
    list_of_regions      rv;
    list_of_cnodes       cv;
    list_of_sysresources sv;
    ;
    list_of_metrics mv;


    bool unset_regions = false;  /// marks, that region was set as "subroutines" and have to be unset after the calculation

    //set all values of all items to 0.0
    initializeValues();
//         cout << "Calculate bsic values for System ! " << endl;
    int index, metricIndex, callIndex, systemIndex;
    getIndices( index, metricIndex, callIndex, systemIndex );


    if ( tabWidget->getMetricTree()->Items().size() == 0 ||
         tabWidget->getCallTree()->Items().size() == 0 ||
         tabWidget->getSystemTree()->Items().size() == 0
         )
    {
        return;
    }
    if ( tabWidget->getMetricTree()->selectedItems().size() == 0 ||
         tabWidget->getCallTree()->selectedItems().size() == 0 ||
         tabWidget->getSystemTree()->selectedItems().size() == 0
         )
    {
        return;
    }

    if ( metricIndex == 0 )
    {
        for ( int i = 0; i < ( int )top.size(); i++ )
        {
            computeValue( top[ i ] );
        }
    }



    if ( metricIndex == 1 && callIndex == 0 )
    {
        if ( tabWidget->getCallTabWidget()->getCurrentCallTree()->getType() == CALLFLAT )
        {
            std::vector<TreeWidgetItem*> sel_regions = tabWidget->getCurrentCallTree()->selectedItems();
            for ( std::vector<TreeWidgetItem*>::iterator riter = sel_regions.begin(); riter != sel_regions.end(); riter++ )
            {
                region_pair rp;
                rp.first =   ( cube::Region* )( ( *riter )->cubeObject );
                if ( rp.first == NULL )
                {
                    unset_regions = true;
                    cube::Region* reg = ( cube::Region* )( ( *riter )->parent()->cubeObject );
                    reg->set_as_subroutines();
                    rp.first = reg;
                }

                rp.second = ( ( *riter )->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
                rv.push_back( rp );
            }
            for ( int i = 0; i < ( int )top.size(); i++ )
            {
                computeValue( rv, top[ i ] );
            }
        }
        else // CALLTREE
        {
            CallTreeWidget*              callTree      = dynamic_cast<CallTreeWidget*> ( tabWidget->getCallTree() );
            std::vector<TreeWidgetItem*> sel_callnodes = callTree->selectedItemsAndIterations();

            for ( std::vector<TreeWidgetItem*>::iterator citer = sel_callnodes.begin(); citer != sel_callnodes.end(); citer++ )
            {
                cnode_pair cp;
                cp.first  = ( cube::Cnode* )( ( *citer )->cubeObject );
                cp.second = ( ( *citer )->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
                cv.push_back( cp );
            }
            for ( int i = 0; i < ( int )top.size(); i++ )
            {
                computeValue( cv, top[ i ] );
            }
        }
    }

    if ( metricIndex == 1 && callIndex == 2 )
    {
        std::vector<TreeWidgetItem*> sel_systemnodes = tabWidget->getSystemTree()->selectedItems();
        for ( std::vector<TreeWidgetItem*>::iterator siter = sel_systemnodes.begin(); siter != sel_systemnodes.end(); siter++ )
        {
            sysres_pair sp;
            sp.first  = ( cube::Sysres* )( ( *siter )->cubeObject );
            sp.second = ( ( *siter )->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
            sv.push_back( sp );
        }
        for ( int i = 0; i < ( int )top.size(); i++ )
        {
            computeValue( sv, top[ i ] );
        }
    }


    if ( metricIndex == 2 )
    {
        std::vector<TreeWidgetItem*> sel_systemnodes = tabWidget->getSystemTree()->selectedItems();
        for ( std::vector<TreeWidgetItem*>::iterator siter = sel_systemnodes.begin(); siter != sel_systemnodes.end(); siter++ )
        {
            sysres_pair sp;
            sp.first  = ( cube::Sysres* )( ( *siter )->cubeObject );
            sp.second = ( ( *siter )->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
            sv.push_back( sp );
        }

        if ( tabWidget->getCallTabWidget()->getCurrentCallTree()->getType() == CALLFLAT )
        {
            std::vector<TreeWidgetItem*> sel_regions = tabWidget->getCurrentCallTree()->selectedItems();
            for ( std::vector<TreeWidgetItem*>::iterator riter = sel_regions.begin(); riter != sel_regions.end(); riter++ )
            {
                region_pair rp;
                rp.first =   ( cube::Region* )( ( *riter )->cubeObject );
                if ( rp.first == NULL )
                {
                    unset_regions = true;
                    cube::Region* reg = ( cube::Region* )( ( *riter )->parent()->cubeObject );
                    reg->set_as_subroutines();
                    rp.first = reg;
                }

                rp.second = ( ( *riter )->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
                rv.push_back( rp );
            }
            for ( int i = 0; i < ( int )top.size(); i++ )
            {
                computeValue( rv, sv, top[ i ] );
            }
        }
        else // CALLTREE
        {
            CallTreeWidget*              callTree      = dynamic_cast<CallTreeWidget*> ( tabWidget->getCallTree() );
            std::vector<TreeWidgetItem*> sel_callnodes = callTree->selectedItemsAndIterations();

            for ( std::vector<TreeWidgetItem*>::iterator citer = sel_callnodes.begin(); citer != sel_callnodes.end(); citer++ )
            {
                cnode_pair cp;
                cp.first  = ( cube::Cnode* )( ( *citer )->cubeObject );
                cp.second = ( ( *citer )->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
                cv.push_back( cp );
            }
            for ( int i = 0; i < ( int )top.size(); i++ )
            {
                computeValue( cv, sv, top[ i ] );
            }
        }
    }

    if ( unset_regions )
    {
        for ( list_of_regions::iterator riter = rv.begin(); riter != rv.end(); riter++ )
        {
            ( *riter ).first->unset_as_subroutines();
        }
    }


    for ( int i = 0; i < ( int )top.size(); i++ )
    {
        validateValue( top[ i ] );
    }
}

QIcon
MetricTreeWidget::icon() const
{
    return QIcon( ":images/tree-icon.png" );
}




//end of computeBasicValues()
