/****************************************************************************
**  CUBE        http://www.score-p.org/                                    **
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2014                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/

#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QPushButton>
#include <QLabel>
#include <QWhatsThis>
#include <QToolTip>
#include <QUrl>
#include <QFileDialog>
#include <QTextStream>
#include <QClipboard>
#include <QApplication>
#include <QMimeData>

#include "NewDerivatedMetricWidget.h"
#include "CubePLSyntaxHighlighter.h"
#include "MessageWidget.h"
#include "HtmlHelpCollection.h"
#include "DerivedMetricsCollection.h"

/*--- Constructors & destructor -------------------------------------------*/

/**
 * @brief Constructs a widget for construction of a new derivated metric
 *
 * This constructor creates a new help browser dialog with the given window
 * @p title and @p parent widget.
 *
 * @param  title   Window title
 * @param  parent  Parent widget
 */
NewDerivatedMetricWidget::NewDerivatedMetricWidget( cube::Cube*   _cube,
                                                    cube::Metric* _met_parent,
                                                    QWidget*      parent )
    : QDialog( parent ), cube( _cube ), parent_metric( _met_parent )
{
    help_browser = NULL;
    wwidget      = new MessageWidget( PermanentError,  QString( "" ) );
    wwidget->setFocusPolicy( Qt::NoFocus );
    wwidget->hide();
    setAcceptDrops( true );
    created_metric      = NULL;
    calculation_ok      = true;
    calculation_init_ok = true;
    if ( parent_metric != NULL )
    {
        uniq_name = parent_metric->get_uniq_name();
        disp_name = parent_metric->get_disp_name();
        uom       = parent_metric->get_uom();
        val       = parent_metric->get_val();
        url       = parent_metric->get_url();
        descr     = parent_metric->get_descr();
        kind      = parent_metric->get_metric_kind();
    }

    setWindowTitle( tr( "Create new metric as a child of metric " ).append( QString::fromStdString( disp_name ) ) );
    QVBoxLayout* top_layout = new QVBoxLayout();
    top_layout->setSpacing( 5 );
    QHBoxLayout* layout1 = new QHBoxLayout();

    layout1->addWidget( new QLabel( tr( "<b>Select metric from collection :</b>" ) ) );
    metric_library_selection = new QComboBox();
    metric_library_selection->addItems( DerivedMetricsCollection::getDerivedMetricsNames() );
    metric_library_selection->setCurrentIndex( 0 );
    metric_library_selection->setWhatsThis( HelpCollection::getHelpText( WHAT_IS_DERIVED_METRICS_LIBRARY ) );
    layout1->addWidget( metric_library_selection );
    connect( metric_library_selection, SIGNAL( currentIndexChanged( int  ) ), this, SLOT( selectMetricFromLibrary( int  ) ) );

    layout1->addSpacing( 10 );

    metric_from_file = new QPushButton( QIcon( ":images/fromtextfile.png" ), "" );
    metric_from_file->setWhatsThis( HelpCollection::getHelpText( WHAT_IS_FROM_FILE ) );
    layout1->addWidget( metric_from_file );
    connect( metric_from_file, SIGNAL( pressed() ), this, SLOT( createMetricFromFile() ) );

    metric_from_clipboard = new QPushButton( QIcon( ":images/paste-icon.png" ), "" );
    metric_from_clipboard->setWhatsThis( HelpCollection::getHelpText( WHAT_IS_CLIPBOARD ) );
    layout1->addWidget( metric_from_clipboard );
    connect( metric_from_clipboard, SIGNAL( pressed() ), this, SLOT( createMetricFromClipboard() ) );

    layout1->addStretch();

    top_layout->addLayout( layout1 );
    QWidget* horizontalLineWidget = new QWidget;
    horizontalLineWidget->setFixedHeight( 2 );
    horizontalLineWidget->setSizePolicy( QSizePolicy::Expanding, QSizePolicy::Fixed );
    horizontalLineWidget->setStyleSheet( QString( "background-color: #c0c0c0;" ) );
    top_layout->addWidget( horizontalLineWidget  );

    QGridLayout* grid_layout2 = new QGridLayout();

    grid_layout2->addWidget( new QLabel( tr( "<b>Derived metric type :</b>" ) ), 0, 0 );
    metric_type_selection = new QComboBox();
    metric_type_selection->addItem( " --- please select type of derived metric ---- " );
    metric_type_selection->addItem( "Postderived metric" );
    metric_type_selection->addItem( "Prederived inclusive metric" );
    metric_type_selection->addItem( "Prederived exclusive metric" );
    metric_type_selection->setCurrentIndex( 0 );
    grid_layout2->addWidget( metric_type_selection, 0, 1 );
    connect( metric_type_selection, SIGNAL( currentIndexChanged( int  ) ), this, SLOT( setMetricType( int  ) ) );


    _metric_kinds_help = new QPushButton( QIcon( ":images/whatishelp.png" ), "" );
    grid_layout2->addWidget( _metric_kinds_help, 0, 2 );
    connect( _metric_kinds_help, SIGNAL( pressed() ), this, SLOT( showKindsOfDerivedMetricsHelp() ) );






    grid_layout2->addWidget( new QLabel( tr( "<b>Display name :</b>" ) ), 1, 0 );
    display_name_input = new QLineEdit();
    grid_layout2->addWidget( display_name_input, 1, 1 );
    connect( display_name_input, SIGNAL( textChanged( const QString & ) ), this, SLOT( setDisplayName( const QString & ) ) );

    grid_layout2->addWidget( new QLabel( tr( "<b>Unique name :</b>" ) ), 2, 0 );
    unique_name_input =  new QLineEdit();
    grid_layout2->addWidget( unique_name_input, 2, 1 );
    connect( unique_name_input, SIGNAL( textChanged( const QString & ) ), this, SLOT( setUniqName( const QString & ) ) );

    grid_layout2->addWidget( new QLabel( tr( "<b>Data type :</b>" ) ), 3, 0 );
    grid_layout2->addWidget(  new QLabel( ( parent_metric == NULL ) ? QString::fromStdString( "DOUBLE" ) : QString::fromStdString( parent_metric->get_dtype() ) ), 3, 1 );

    grid_layout2->addWidget( new QLabel( tr( "<b>Unit of measurement :</b>" ) ), 4, 0 );
    uom_input = new QLineEdit();
    grid_layout2->addWidget( uom_input, 4, 1 );
    connect( uom_input, SIGNAL( textChanged( const QString & ) ), this, SLOT( setUom( const QString & ) ) );

    grid_layout2->addWidget( new QLabel( tr( "<b>URL :</b>" ) ), 5, 0 );
    url_input = new QLineEdit();
    grid_layout2->addWidget( url_input, 5, 1 );
    connect( url_input, SIGNAL( textChanged( const QString & ) ), this, SLOT( setURL( const QString & ) ) );

    top_layout->addLayout( grid_layout2 );
    QVBoxLayout* layout3 = new QVBoxLayout();


    layout3->addWidget( new QLabel( tr( "<b>Description :</b>" ) ) );
    description_input = new QTextEdit();
    layout3->addWidget( description_input );
    connect( description_input, SIGNAL( textChanged() ), this, SLOT( setDescription() ) );


    top_layout->addLayout( layout3 );

    QGridLayout* exp_layout = new QGridLayout();

    QHBoxLayout* grid_layout_exp = new QHBoxLayout();
    grid_layout_exp->addWidget( new QLabel( tr( "<b>Calculation (CubePL) :</b>" ) ), Qt::AlignLeft );

    _metric_expt_help = new QPushButton( QIcon( ":images/whatishelp.png" ), "" );
    grid_layout_exp->addWidget( _metric_expt_help, Qt::AlignLeft );
    connect( _metric_expt_help, SIGNAL( pressed() ), this, SLOT( showMetricExpressionHelp() ) );
    grid_layout_exp->addStretch( 10 );

    exp_layout->addLayout( grid_layout_exp, 0, 0 );



    calculation_input = new QTextEdit();
    exp_layout->addWidget( calculation_input, 1, 0 );
    calculation_input->setStyleSheet( "color:black; background: white; selection-color: red;" );
    connect( calculation_input, SIGNAL( textChanged() ), this, SLOT( setCubePLExpression() ) );
    connect( calculation_input, SIGNAL( cursorPositionChanged() ), this, SLOT( showAvailableMetrics() ) );
    error_highlighter = new CubePLSyntaxHighlighter( calculation_input->document() );

    QHBoxLayout* grid_layout_init_exp = new QHBoxLayout();
    grid_layout_init_exp->addWidget( new QLabel( tr( "<b>Initialization (CubePL) :</b>" ) ), Qt::AlignLeft  );

    _metric_init_expt_help = new QPushButton( QIcon( ":images/whatishelp.png" ), "" );
    grid_layout_init_exp->addWidget( _metric_init_expt_help, Qt::AlignLeft );
    connect( _metric_init_expt_help, SIGNAL( pressed() ), this, SLOT( showMetricInitExpressionHelp() ) );
    grid_layout_init_exp->addStretch( 10 );



    exp_layout->addLayout( grid_layout_init_exp, 0, 1 );


    calculation_init_input = new QTextEdit();
    exp_layout->addWidget( calculation_init_input, 1, 1 );
    calculation_init_input->setStyleSheet( "color:black; background: white; selection-color: red;" );
    connect( calculation_init_input, SIGNAL( textChanged() ), this, SLOT( setCubePLInitExpression() ) );
    connect( calculation_init_input, SIGNAL( cursorPositionChanged() ), this, SLOT( showAvailableMetrics() ) );
    error_init_highlighter = new CubePLSyntaxHighlighter( calculation_init_input->document() );

    _whatis_help = new QPushButton( QIcon( ":images/whatishelp.png" ), "" );

    exp_layout->addWidget( _whatis_help, 1, 2, Qt::AlignTop );
    connect( _whatis_help, SIGNAL( pressed() ), this, SLOT( showCubePLHelp() ) );

    QHBoxLayout* syntax_status_layout      = new QHBoxLayout();
    QHBoxLayout* syntax_init_status_layout = new QHBoxLayout();


    syntax_status_layout->addWidget( new QLabel( tr( "<b>Syntax check :</b>" ) ) );
    syntax_status = new QLabel( "" );
    syntax_status_layout->addWidget( syntax_status,  Qt::AlignLeft );

    syntax_init_status_layout->addWidget( new QLabel( tr( "<b>Init syntax check :</b>" ) ) );
    init_syntax_status = new QLabel( "" );
    syntax_init_status_layout->addWidget( init_syntax_status,  Qt::AlignLeft );


    exp_layout->addLayout( syntax_status_layout, 2, 0 );
    exp_layout->addLayout( syntax_init_status_layout, 2, 1 );
    top_layout->addLayout( exp_layout );

    QHBoxLayout* button_layout = new  QHBoxLayout();

    create_metric = new QPushButton( tr( "Create &metric" ) );
    connect( create_metric, SIGNAL( pressed() ), this, SLOT( createMetric() ) );
    button_layout->addWidget( create_metric );
    create_metric->setEnabled( unique_name_input->text() != "" );

    QPushButton* _close = new QPushButton( tr( "&Cancel" ) );
    connect( _close, SIGNAL( pressed() ), this, SLOT( close() ) );
    button_layout->addWidget( _close );



    top_layout->addLayout( button_layout );
    QWidget* horizontalLineWidget2 = new QWidget;
    horizontalLineWidget2->setFixedHeight( 2 );
    horizontalLineWidget2->setSizePolicy( QSizePolicy::Expanding, QSizePolicy::Fixed );
    horizontalLineWidget2->setStyleSheet( QString( "background-color: #c0c0c0;" ) );
    top_layout->addWidget( horizontalLineWidget2  );

    share_metric = new QLabel( ""  );
    enableShareLink( false );
    top_layout->addWidget( share_metric );

    setLayout( top_layout );
}

void
NewDerivatedMetricWidget::setMetricType(  int type )
{
    switch ( type )
    {
        case 1:
            metric_type = cube::CUBE_METRIC_POSTDERIVED;
            break;
        case 2:
            metric_type = cube::CUBE_METRIC_PREDERIVED_INCLUSIVE;
            break;
        case 3:
            metric_type = cube::CUBE_METRIC_PREDERIVED_EXCLUSIVE;
            break;
        default:
            metric_type = cube::CUBE_METRIC_SIMPLE; // just value to set the widget in unenabled state
            break;
    }
    create_metric->setEnabled(  ( metric_type == cube::CUBE_METRIC_POSTDERIVED || metric_type == cube::CUBE_METRIC_PREDERIVED_EXCLUSIVE || metric_type == cube::CUBE_METRIC_PREDERIVED_INCLUSIVE   )
                                && ( disp_name != "" )
                                && ( uniq_name != "" )
                                && ( calculation != "" )
                                && ( calculation_ok )
                                && ( calculation_init_ok )
                                );
}

void
NewDerivatedMetricWidget::setDisplayName(  const QString& new_disp_name )
{
    disp_name = new_disp_name.toStdString();
    create_metric->setEnabled(  ( metric_type == cube::CUBE_METRIC_POSTDERIVED || metric_type == cube::CUBE_METRIC_PREDERIVED_EXCLUSIVE || metric_type == cube::CUBE_METRIC_PREDERIVED_INCLUSIVE   )
                                && ( disp_name != "" )
                                && ( uniq_name != "" )
                                && ( calculation != "" )
                                && ( calculation_ok )
                                && ( calculation_init_ok )
                                );
}


void
NewDerivatedMetricWidget::setUniqName( const QString& new_uniq_name )
{
    uniq_name = new_uniq_name.toStdString();
    create_metric->setEnabled(  ( metric_type == cube::CUBE_METRIC_POSTDERIVED || metric_type == cube::CUBE_METRIC_PREDERIVED_EXCLUSIVE || metric_type == cube::CUBE_METRIC_PREDERIVED_INCLUSIVE   )
                                && ( disp_name != "" )
                                && ( uniq_name != "" )
                                && ( calculation != "" )
                                && ( calculation_ok )
                                && ( calculation_init_ok )
                                );
}


void
NewDerivatedMetricWidget::setUom( const QString& new_uom )
{
    uom = new_uom.toStdString();
}

void
NewDerivatedMetricWidget::setValue( const QString& new_value )
{
    val = new_value.toStdString();
}

void
NewDerivatedMetricWidget::setURL( const QString& new_url )
{
    url = new_url.toStdString();
}

void
NewDerivatedMetricWidget::setDescription()
{
    descr = description_input->toPlainText().toStdString();
}

void
NewDerivatedMetricWidget::setCubePLExpression()
{
    calculation = calculation_input->toPlainText().toStdString();

    std::string _to_test = std::string( "<cubepl>" ) + calculation + std::string( "</cubepl>" );

    std::string error_message = "";
    if ( cube->test_cubepl_expression( _to_test, error_message ) )
    {
        // OK
//      QPen pen;
//      pen.setColor("green");
//      syntax_status->setPen(pen);
        syntax_status->setStyleSheet( "QLabel { color: green;  }" );
        syntax_status->setText( "OK" );
        wwidget->hide();
        error_highlighter->setErrorPlace( -1 );
        calculation_ok = true;
    }
    else
    {
        // syntax error
        syntax_status->setStyleSheet( "QLabel { color: red;  }" );

        QString error_string = QString::fromStdString( error_message );


        syntax_status->setText( QString( "Syntax error!" ) );

        wwidget->setMessage( PermanentError, error_string );
        wwidget->setFocusPolicy( Qt::NoFocus );
        wwidget->show();


        QStringList list1        = error_string.split( ":" );
        QString     locations    = list1.at( 0 );
        QStringList list2        = locations.split( "." );
        QString     row          = list2.at( 0 );
        QString     columns      = ( list2.size() > 1 ) ? list2.at( 1 ) : row;
        QStringList list3        = columns.split( "-" );
        QString     column_start = list3.at( 0 );

        QString column_end = ( list3.size() > 1 ) ? list3.at( 1 ) : column_start;



//         int errorplace    = column_start.toInt();
        int errorplace_to = column_end.toInt();
        error_highlighter->setErrorPlace( errorplace_to );
        calculation_ok = false;
    }

    create_metric->setEnabled(  ( metric_type == cube::CUBE_METRIC_POSTDERIVED || metric_type == cube::CUBE_METRIC_PREDERIVED_EXCLUSIVE || metric_type == cube::CUBE_METRIC_PREDERIVED_INCLUSIVE   )
                                && ( disp_name != "" )
                                && ( uniq_name != "" )
                                && ( calculation != "" )
                                && ( calculation_ok )
                                && ( calculation_init_ok )
                                );
    enableShareLink( create_metric->isEnabled() );
}


void
NewDerivatedMetricWidget::setCubePLInitExpression()
{
    init_calculation = calculation_init_input->toPlainText().toStdString();

    std::string _to_test = std::string( "<cubepl>" ) + init_calculation + std::string( "</cubepl>" );

    std::string error_message = "";
    if ( cube->test_cubepl_expression( _to_test, error_message ) )
    {
        // OK
        init_syntax_status->setStyleSheet( "QLabel { color: green;  }" );
        init_syntax_status->setText( "OK" );
        wwidget->hide();
        error_init_highlighter->setErrorPlace( -1 );
        calculation_init_ok = true;
    }
    else
    {
        // syntax error
        init_syntax_status->setStyleSheet( "QLabel { color: red;  }" );

        QString error_string = QString::fromStdString( error_message );


        init_syntax_status->setText( QString( "Syntax error!" ) );

        wwidget->setMessage( PermanentError, error_string );
        wwidget->setFocusPolicy( Qt::NoFocus );
        wwidget->show();


        QStringList list1        = error_string.split( ":" );
        QString     locations    = list1.at( 0 );
        QStringList list2        = locations.split( "." );
        QString     row          = list2.at( 0 );
        QString     columns      = ( list2.size() > 1 ) ? list2.at( 1 ) : row;
        QStringList list3        = columns.split( "-" );
        QString     column_start = list3.at( 0 );

        QString column_end = ( list3.size() > 1 ) ? list3.at( 1 ) : column_start;



//         int errorplace    = column_start.toInt();
        int errorplace_to = column_end.toInt();
        error_init_highlighter->setErrorPlace( errorplace_to );
        calculation_init_ok = false;
    }

    create_metric->setEnabled(  ( metric_type == cube::CUBE_METRIC_POSTDERIVED || metric_type == cube::CUBE_METRIC_PREDERIVED_EXCLUSIVE || metric_type == cube::CUBE_METRIC_PREDERIVED_INCLUSIVE   )
                                && ( disp_name != "" )
                                && ( uniq_name != "" )
                                && ( calculation != "" )
                                && ( calculation_ok )
                                && ( calculation_init_ok )
                                );
    enableShareLink( create_metric->isEnabled() );
}





void
NewDerivatedMetricWidget::createMetric()
{
    created_metric = cube->def_met(
        disp_name, uniq_name,
        ( parent_metric == NULL ) ? "DOUBLE" :  parent_metric->get_dtype(),
        uom,
        val,
        url,
        descr,
        parent_metric,
        metric_type,
        calculation,
        init_calculation );

    if ( created_metric == NULL )
    {
        MessageWidget* wwidget = new MessageWidget( Error, QString::fromStdString( "Cannot create metric. Please check the type of parent metric." ) );
        wwidget->show();
    }
    else
    {
        close();
    }
}




void
NewDerivatedMetricWidget::showCubePLHelp()
{
    if ( help_browser == NULL )
    {
        help_browser = new HelpBrowser( tr( "Help" ), this );
    }
    help_browser->showHelpAsString( HelpCollection::getHelpText( CUBEPL_HELP ) );
    QApplication::restoreOverrideCursor();
}

void
NewDerivatedMetricWidget::showKindsOfDerivedMetricsHelp()
{
    if ( help_browser == NULL )
    {
        help_browser = new HelpBrowser( tr( "Help" ), this );
    }
    help_browser->showHelpAsString( HelpCollection::getHelpText( DERIVED_METRIC_HELP ) );
    QApplication::restoreOverrideCursor();
}


void
NewDerivatedMetricWidget::showMetricInitExpressionHelp()
{
    if ( help_browser == NULL )
    {
        help_browser = new HelpBrowser( tr( "Help" ), this );
    }
    help_browser->showHelpAsString( HelpCollection::getHelpText( DERIVED_METRIC_INIT_EXPRESSION_HELP ) );
    QApplication::restoreOverrideCursor();
}

void
NewDerivatedMetricWidget::showMetricExpressionHelp()
{
    if ( help_browser == NULL )
    {
        help_browser = new HelpBrowser( tr( "Help" ), this );
    }
    help_browser->showHelpAsString( HelpCollection::getHelpText( DERIVED_METRIC_EXPRESSION_HELP  ) );
    QApplication::restoreOverrideCursor();
}




void
NewDerivatedMetricWidget::showAvailableMetrics()
{
#if QT_VERSION >= 0x040500

    QTextDocument* _document        = calculation_input->document();
    QTextCursor    _current_cursor  = calculation_input->textCursor();
    int            _cursor_position = _current_cursor.position();
    if (
        _document->characterAt( _cursor_position - 1 ).toLatin1() == ':' &&
        _document->characterAt( _cursor_position - 2 ).toLatin1() == ':' &&
        _document->characterAt( _cursor_position - 3 ).toLatin1() == 'c' &&
        _document->characterAt( _cursor_position - 4 ).toLatin1() == 'i' &&
        _document->characterAt( _cursor_position - 5 ).toLatin1() == 'r' &&
        _document->characterAt( _cursor_position - 6 ).toLatin1() == 't' &&
        _document->characterAt( _cursor_position - 7 ).toLatin1() == 'e' &&
        _document->characterAt( _cursor_position - 8 ).toLatin1() == 'm'
        )
    {
        vector<cube::Metric*> _metrics           = cube->get_metv();
        QString               _available_metrics =
            "<h3>Available metrics</h3>"
            "<table style=\"color: white;\" border= \"5px\" bgcolor=\"#333333\" cellpadding=\"10px\" ><tr><th><u>    Display name</u>  </th>  <th>    <u>Unique name</u> </th></tr>\n";
        for ( vector<cube::Metric*>::iterator m_iter = _metrics.begin(); m_iter != _metrics.end(); m_iter++ )
        {
            _available_metrics.append( QString::fromStdString(
                                           "<tr><td>" + ( *m_iter )->get_disp_name() + " </td><td>  " + ( *m_iter )->get_uniq_name() + "</td></tr>\n" ) );
        }
        _available_metrics.append( QString::fromStdString( "</table>\n" ) );
        QPoint _point1 = this->mapToGlobal( rect().bottomRight() );
        QPoint _point2 = calculation_input->mapToGlobal( QPoint() );
        QPoint _target_point( _point1.x(), _point2.y() );

        QToolTip::showText( _target_point, _available_metrics, this, rect() );
    }
    else
    {
        QToolTip::hideText();
    }
#endif
}



void
NewDerivatedMetricWidget::dragEnterEvent( QDragEnterEvent* event )
{
    if ( event->mimeData()->hasUrls() )
    {
        event->acceptProposedAction();
    }
}


void
NewDerivatedMetricWidget::dropEvent( QDropEvent* event )
{
    QList<QUrl>           _urls = event->mimeData()->urls();
    QList<QUrl>::iterator _iter = _urls.begin();
    QUrl                  _file = *_iter;
    event->acceptProposedAction();
    fillTheFormFromUrl( _file );
}



void
NewDerivatedMetricWidget::selectFileForFillingForm()
{
    QString _file = QFileDialog::getOpenFileName( this, tr( "Open File with definition of a derived metric" ),
                                                  "/home",
                                                  tr( "Text files(*.txt *.dm )" ) );
    if ( _file != "" )
    {
        QUrl _url( _file );
        fillTheFormFromUrl( _url );
    }
}

void
NewDerivatedMetricWidget::fillTheFormFromUrl( QUrl& _url )
{
    QFile _file( _url.path() );
    if ( !_file.open( QIODevice::ReadOnly | QIODevice::Text ) )
    {
        return;
    }

    QTextStream in( &_file );
    fillTheForm( in );
}


void
NewDerivatedMetricWidget::fillTheFormFromClipboard()
{
    const QClipboard* clipboard = QApplication::clipboard();
    const QMimeData*  mimeData  = clipboard->mimeData();

    if ( mimeData->hasText() )
    {
        QString     text = mimeData->text();
        QTextStream in( &text );
        fillTheForm( in );
    }
}




void
NewDerivatedMetricWidget::fillTheForm( QTextStream& in )
{
    enum TargetString
    {
        DM_PARSER_COMMENT         = -1,
        DM_PARSER_TYPE            = 0,
        DM_PARSER_DISPLAY_NAME    = 1,
        DM_PARSER_UNIQUE_NAME     = 2,
        DM_PARSER_UOM             = 3,
        DM_PARSER_URL             = 4,
        DM_PARSER_DESCRIPTION     = 5,
        DM_PARSER_EXPRESSION      = 6,
        DM_PARSER_INIT_EXPRESSION = 7
    };

    TargetString switcher = DM_PARSER_COMMENT;

    QRegExp dmetric_commentar( "^\\s*#" );
    QRegExp dmetric_type( "^\\s*metric type\\s*:" );
    QRegExp dmetric_display_name( "^\\s*display name\\s*:" );
    QRegExp dmetric_unique_name( "^\\s*unique name\\s*:" );
    QRegExp dmetric_uom( "^\\s*uom\\s*:" );
    QRegExp dmetric_url( "^\\s*url\\s*:" );
    QRegExp dmetric_description( "^\\s*description\\s*:" );
    QRegExp dmetric_expression( "^\\s*cubepl expression\\s*:" );
    QRegExp dmetric_init_expression( "^\\s*cubepl init expression\\s*:" );

    QString metric_type;
    QString metric_display_name;
    QString metric_unique_name;
    QString metric_uom;
    QString metric_url;
    QString metric_description;
    QString metric_cubepl_expression;
    QString metric_cubepl_init_expression;

    QString parts[ 8 ];

    while ( !in.atEnd() )
    {
        QString line = in.readLine().trimmed().toLower();

        if ( line.indexOf( dmetric_commentar ) != -1 )
        {
            continue;
        }
        else if ( line.indexOf( dmetric_type ) != -1 )
        {
            switcher = DM_PARSER_TYPE;
            line.remove( dmetric_type );
        }
        else if ( line.indexOf( dmetric_display_name ) != -1 )
        {
            switcher = DM_PARSER_DISPLAY_NAME;
            line.remove( dmetric_display_name );
        }
        else if ( line.indexOf( dmetric_unique_name ) != -1 )
        {
            switcher = DM_PARSER_UNIQUE_NAME;
            line.remove( dmetric_unique_name );
        }
        else if ( line.indexOf( dmetric_uom ) != -1 )
        {
            switcher = DM_PARSER_UOM;
            line.remove( dmetric_uom );
        }
        else if ( line.indexOf( dmetric_url ) != -1 )
        {
            switcher = DM_PARSER_URL;
            line.remove( dmetric_url );
        }
        else if ( line.indexOf( dmetric_description ) != -1 )
        {
            switcher = DM_PARSER_DESCRIPTION;
            line.remove( dmetric_description );
        }
        else if ( line.indexOf( dmetric_expression ) != -1 )
        {
            switcher = DM_PARSER_EXPRESSION;
            line.remove( dmetric_expression );
        }
        else if ( line.indexOf( dmetric_init_expression ) != -1 )
        {
            switcher = DM_PARSER_INIT_EXPRESSION;
            line.remove( dmetric_init_expression );
        }

        if ( switcher != DM_PARSER_COMMENT )
        {
            parts[ switcher ].append( line ).append( "\n" );
        }
    }

    if ( parts[ DM_PARSER_TYPE ].trimmed().toLower() == "postderived" )
    {
        metric_type_selection->setCurrentIndex( 1 );
    }
    else
    if ( parts[ DM_PARSER_TYPE ].trimmed().toLower() == "prederived_inclusive" )
    {
        metric_type_selection->setCurrentIndex( 2 );
    }
    else
    if ( parts[ DM_PARSER_TYPE ].trimmed().toLower() == "prederived_exclusive" )
    {
        metric_type_selection->setCurrentIndex( 3 );
    }
    else
    {
        metric_type_selection->setCurrentIndex( 0 );
    }

    display_name_input->setText( parts[ DM_PARSER_DISPLAY_NAME ].trimmed() );
    unique_name_input->setText( parts[ DM_PARSER_UNIQUE_NAME ].trimmed() );
    uom_input->setText( parts[ DM_PARSER_UOM ].trimmed() );
    url_input->setText( parts[ DM_PARSER_URL ].trimmed() );
    description_input->setText( parts[ DM_PARSER_DESCRIPTION ].trimmed() );
    calculation_input->setText( parts[ DM_PARSER_EXPRESSION ].trimmed() );
    calculation_init_input->setText( parts[ DM_PARSER_INIT_EXPRESSION ].trimmed() );
}





void
NewDerivatedMetricWidget::createMetricFromFile()
{
    selectFileForFillingForm();
}


void
NewDerivatedMetricWidget::createMetricFromClipboard()
{
    fillTheFormFromClipboard();
}


void
NewDerivatedMetricWidget::selectMetricFromLibrary(  int type )
{
    QTextStream in( &DerivedMetricsCollection::getDerivedMetricText( type ) );
    fillTheForm( in );
}



void
NewDerivatedMetricWidget::enableShareLink( bool flag )
{
    QString _caption( tr( "Share this metric with SCALASCA group" ) );
    if ( !flag )
    {
        share_metric->setText( _caption );
        share_metric->setOpenExternalLinks( false  );
    }
    else
    {
        QString _text_intro( tr( "Dear Scalasca Team, \n\n"
                                 "this metric is very useful in my performance analysis and I would like to suggest to include it into the library of Cube:\n\n" ) );
        QString _metric = packDataToString();
        QString _text_closure( tr( "\n\n\n\n\n Sincerely," ) );
        share_metric->setText( tr( "<a href=\"mailto:scalasca@fz-juelich.de?subject=Please add this derived metric to your library&body=" ).
                               append( _text_intro ).
                               append( _metric ).
                               append( _text_closure ).
                               append( "\">" ).
                               append( _caption ).
                               append( "</a> " ) );
        share_metric->setOpenExternalLinks( true );
    }
}


QString
NewDerivatedMetricWidget::packDataToString()
{
    QString _to_return( tr( "Metric type:" ).append( metric_type_selection->currentText().append( "\n\n" ).append(
                                                         tr( "Display name:" ).append( display_name_input->text().append( "\n\n" ).append(
                                                                                           tr( "Unique name:" ).append( unique_name_input->text().append( "\n\n" ).append(
                                                                                                                            tr( "UoM:" ).append( uom_input->text().append( "\n\n" ).append(
                                                                                                                                                     tr( "URL:" ).append( url_input->text().append( "\n\n" ).append(
                                                                                                                                                                              tr( "Description:" ).append( description_input->toPlainText().append( "\n\n" ).append(
                                                                                                                                                                                                               tr( "CubePL Expression:" ).append( calculation_input->toPlainText().append( "\n\n" ).append(
                                                                                                                                                                                                                                                      tr( "CubePl Init Expression:" ).append( calculation_init_input->toPlainText().append( "\n\n" ) ) ) ) ) ) ) ) ) ) ) ) ) ) ) ) );
    _to_return.replace( "\"", "'" );
    return _to_return;
}
