/****************************************************************************
**  CUBE        http://www.score-p.org/                                    **
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2014                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/

#include "SystemBoxPlot.h"
#include <qdebug.h>
#include <algorithm>
#include <cmath>

using namespace std;


/**
   creates a bot plot which displays the values given by TreeWidget
 */
SystemBoxPlot::SystemBoxPlot( QWidget*          parent,
                              SystemTreeWidget* t ) :
    BoxPlot( NULL, t->getTabWidget()->getPrecisionWidget() )
{
    this->systemTreeWidget = t;
    this->valueWidget      = t->getTabWidget()->getValueWidget();
    this->subsetCombo      = NULL;

    setWhatsThis(
        "This Boxplot shows the distribution of values in the system tree."
        " It starts with the smallest value (minimum) and ends with the largest value (maximum)."
        " The bottom and top of the box represent the lower quartile (Q1) and the upper quartile (Q3)."
        " The bold line represents the median (Q2) and the dashed line represents the mean value."
        " To zoom into the boxplot, select an area with the mouse."
        " To zoom out, click with the middle mouse button in the drawing."
        );

    scrollArea = new ScrollArea( parent );
    scrollArea->setWidget( this );
    scrollArea->setWidgetResizable( true );
    scrollArea->setVerticalScrollBarPolicy( Qt::ScrollBarAlwaysOff );
    scrollArea->setHorizontalScrollBarPolicy( Qt::ScrollBarAlwaysOff );

    container = new SplitterContainer();
    container->addWidget( scrollArea );

    // set lower splitter element to minimum size (1 pixel => replaced by minimumSize())
    QList<int> sizeList;
    sizeList << container->size().height() << 1;
    container->setSizes( sizeList );
}

QWidget*
SystemBoxPlot::widget()
{
    return container;
}

void
SystemBoxPlot::valuesChanged()
{
    Item item = calculateStatistics( false );
    Item abs  = calculateStatistics( true );

    double start;
    if ( systemTreeWidget->tabWidget->getValueModus() == ABSOLUTE )
    {
        start = item.min;
    }
    else
    {
        start = 0.0;
    }

    // fill box plot with calculated data
    this->Set( item, abs, start );

    updateValueWidget();
}

QIcon
SystemBoxPlot::icon() const
{
    return QIcon( ":images/boxplot-icon.png" );
}

void
SystemBoxPlot::setSubsetCombo( QComboBox* combo )
{
    subsetCombo = combo;
}

void
SystemBoxPlot::setActive()
{
    if ( !subsetCombo )
    {
        return;
    }
    QSplitter* oldParent = ( QSplitter* )subsetCombo->parentWidget();
    QSplitter* newParent = container;
    QList<int> sizeList  = oldParent->sizes();

    // share the same instance of QCombobox between BoxPlot and SystemTree -> change the parent widget
    subsetCombo->setParent( newParent );

    // save collapsed state
    newParent->setSizes( sizeList );
}

// calculate statistical values
Item
SystemBoxPlot::calculateStatistics( bool absolute )
{
    int    theCount;
    double theSum;
    double theMean;
    double theVariance;
    double theMin;
    double theMax;
    double theMedian;
    double theQ25;
    double theQ75;

    vector<TreeWidgetItem*> items = systemTreeWidget->getActiveSubset();
    vector<double>          values;

    theSum = 0.;
    double value;
    bool   ok;
    for ( unsigned i = 0; i < items.size(); i++ )
    {
        TreeWidgetItem* item = items.at( i );
        // single-threaded processes or thread
        if ( item->childCount() == 0 )
        {
            QString str;
            if ( absolute )
            {
                systemTreeWidget->getValue( item, item->isExpanded(), value, str, ok, ABSOLUTE );
            }
            else
            {
                ValueModus valueModus = systemTreeWidget->getTabWidget()->getValueModus();
                systemTreeWidget->getValue( item, item->isExpanded(), value, str, ok, valueModus );
                //value = item->getValue();
            }
            values.push_back( value );
            theSum += value;
        }
    }

    theCount = values.size();

    if ( theCount == 0 )
    {
        theCount = 0;
        this->Reset();
        return Item( "", 0, 0, 0, 0, 0, 0, 0, 0, 0 ); // not enough values to display
    }
    theMean = theSum / values.size();

    theVariance = 0.;
    for ( unsigned i = 0; i < values.size(); i++ )
    {
        theVariance += ( values.at( i ) - theMean ) * ( values.at( i ) - theMean );
    }
    if ( values.size() > 0 )
    {
        theVariance /= ( values.size() );
    }

    sort( values.begin(), values.end() );
    theMin    = *values.begin();
    theMax    = *( values.end() - 1 );
    theMedian = *( values.begin() + values.size() / 2 );
    theQ25    = *( values.begin() + values.size() / 4 );
    theQ75    = *( values.begin() + values.size() * 3 / 4 );

    return Item( "", theCount, theSum, theMean, theMin, theQ25, theMedian, theQ75, theMax, theVariance );
}

/**
   updates the value widget below the tab widget
 */
void
SystemBoxPlot::updateValueWidget()
{
    vector<double> values;
    Item           item;

    if ( systemTreeWidget->getTabWidget()->getValueModus() == ABSOLUTE )
    {
        item = calculateStatistics( true );
    }
    else
    {
        item = calculateStatistics( false );
    }

    /* from SystemTopologyWidget */
    //in case the topology is empty or if no items are selected, just
    //clear the value widget
    if ( item.count == 0 )
    {
        valueWidget->clear();
    }
    else
    {
        double sigma   = sqrt( item.variance );
        bool   intType = systemTreeWidget->hasIntegerType();
        double displayedValue;

        if ( systemTreeWidget->getTabWidget()->getValueModus() == ABSOLUTE )
        {
            displayedValue = item.mean;
            // in the absolute value mode display the mean value (with the min/max values)
            valueWidget->update( item.min, item.max, displayedValue, intType, false, item.mean, sigma );
        }
        else
        {
            displayedValue = 100; // show +/- sigma
            //in value modes other that the absolute value mode display the
            //current minvalue/value/maxvalue plus absolute
            //minvalue/value/maxvalue
            valueWidget->update( 0.0, 100.0, displayedValue,
                                 item.min, item.max, displayedValue, intType, false, item.mean, sigma );
        }
    }
}
