/****************************************************************************
**  CUBE        http://www.score-p.org/                                    **
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2014                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include "SystemTopologyWidget.h"

#include "SystemTreeWidget.h"
#include "ScrollArea.h"
#include "ValueWidget.h"

#include "CubeCartesian.h"

#include <QScrollBar>
#include <QSettings>
#include <qdebug.h>

#include "SystemTopologyView.h"
#include "SystemTopologyDrawing.h"
//#include "SystemTopologyGL.h"
#include "SystemTopologyToolBar.h"

/**
   controls several widgets:
   TopologyDimensionBar which allows to reorder the dimensions
   SystemTopologyView which draws the topology
 */

/*************** constructor / destructor *************************/

SystemTopologyWidget::SystemTopologyWidget( SystemTreeWidget* systemTreeWidget,
                                            unsigned          topologyId )
{
    SystemTopologyToolBar::getInstance()->reinit();
    data      = new SystemTopologyData( systemTreeWidget, topologyId );
    transform = new SystemTopologyViewTransform( data );
    view      = new SystemTopologyDrawing( data, transform );
    //view = new SystemTopologyGL(data);

    connect( view, SIGNAL( selectItem( int, bool ) ), this, SLOT( selectItem( int, bool ) ) );
    connect( view, SIGNAL( scrollTo( int, int ) ), this, SLOT( scrollTo( int, int ) ) );

    connect( data, SIGNAL( dataChanged() ), view, SLOT( updateDrawing() ) );
    connect( data, SIGNAL( viewChanged() ), view, SLOT( updateDrawing() ) );
    connect( data, SIGNAL( rescaleRequest() ), transform, SLOT( rescale() ) );

    connect( transform, SIGNAL( rescaleRequest() ), view, SLOT( rescaleDrawing() ) );
    connect( transform, SIGNAL( viewChanged() ), view, SLOT( updateDrawing() ) );
    connect( transform, SIGNAL( zoomChanged( double ) ), view, SLOT( changeZoom( double ) ) );
    connect( transform, SIGNAL( xAngleChanged( int ) ), view, SLOT( setXAngle( int ) ) );
    connect( transform, SIGNAL( yAngleChanged( int ) ), view, SLOT( setYAngle( int ) ) );
    connect( transform, SIGNAL( positionChanged( int, int ) ), view, SLOT( move( int, int ) ) );

    connect( systemTreeWidget, SIGNAL( selectionChanged( TreeWidgetItem* ) ), this, SLOT( handleSelection( TreeWidgetItem* ) ) );

    dimensionBar           = 0;
    this->systemTreeWidget = systemTreeWidget;
    this->topologyId       = topologyId;
    this->setWidget( view );
    setHorizontalScrollBarPolicy( Qt::ScrollBarAlwaysOn );
    setVerticalScrollBarPolicy( Qt::ScrollBarAlwaysOn );

    QScrollBar* sbh = horizontalScrollBar();
    QScrollBar* sbv = verticalScrollBar();
    connect( sbh, SIGNAL( valueChanged( int ) ), this, SLOT( hscroll( int ) ) );
    connect( sbv, SIGNAL( valueChanged( int ) ), this, SLOT( vscroll( int ) ) );

    tabWidget = new SplitterContainer();
}

/**
 * @brief returns the widget that will be placed in SystemTabWidget
 */
QWidget*
SystemTopologyWidget::widget()
{
    return tabWidget;
}

void
SystemTopologyWidget::vscroll( int val )
{
    view->scrolledTo( horizontalScrollBar()->value(), val );
}

void
SystemTopologyWidget::hscroll( int val )
{
    view->scrolledTo( val, verticalScrollBar()->value() );
}

SystemTopologyWidget::~SystemTopologyWidget()
{
}

QIcon
SystemTopologyWidget::icon() const
{
    return QIcon( ":images/topology-icon.png" );
}

QString
SystemTopologyWidget::label() const
{
    return topologyName;
}

void
SystemTopologyWidget::resizeEvent( QResizeEvent* )
{
    view->setSize( QSize( viewport()->width(), viewport()->height() ) );
}

QSize
SystemTopologyWidget::sizeHint() const
{
    return view->sizeHint();
}

/* initializes the topology with data from the cube object
 */
void
SystemTopologyWidget::initialize( cube::Cube* cube )
{
    selectMode = FOLD;
    TopologyDimensionBar* bar = ( TopologyDimensionBar* )getDimensionSelectionBar( cube );
    data->setFoldingDimensions( bar->getFoldingVector() );
    data->initialize( cube );

    topologyName = ( cube->get_cartv() ).at( topologyId )->get_name().c_str();
    if ( topologyName == "" )
    {
        topologyName.append( "Topology " );
        topologyName.append( QString::number( topologyId ) );
    }

    // --- generate widget to be displayed in SystemTab -----
    tabWidget->addWidget( this );
    /** add topology dimension toolbar with scrollPane */
    QWidget* dimBar = getDimensionSelectionBar( cube );
    if ( dimBar != 0 )
    {
        QScrollArea* scroll = new QScrollArea();
        tabWidget->addWidget( scroll );
        scroll->setVerticalScrollBarPolicy( Qt::ScrollBarAlwaysOff );
        scroll->setHorizontalScrollBarPolicy( Qt::ScrollBarAlwaysOff );
        scroll->setFrameStyle( QFrame::NoFrame );
        scroll->setMinimumHeight( dimBar->minimumSizeHint().height() );
        scroll->setMaximumHeight( dimBar->minimumSizeHint().height() );
        scroll->setWidget( dimBar );
        long ndims = ( cube->get_cartv() ).at( this->topologyId )->get_ndims();
        if ( ndims <= 3 ) // minimize dimension selection bar
        {
            QList<int> sizeList;
            sizeList << 1 << 0;
            tabWidget->setSizes( sizeList );
        }
    }
}


/**************************** miscellaneous **********************/


/** create toolbar to select dimensions to display
 */
QWidget*
SystemTopologyWidget::getDimensionSelectionBar( cube::Cube* cube )
{
    if ( dimensionBar == 0 )
    {
        const std::vector<long>&       dims  = cube->get_cart( topologyId )->get_dimv();
        const std::vector<std::string> names = cube->get_cart( topologyId )->get_namedims();
        dimensionBar = new TopologyDimensionBar( dims, names );
        connect( dimensionBar, SIGNAL( selectedDimensionsChanged( std::vector<long>) ), data, SLOT( selectedDimensionsChanged( std::vector<long>) ) );
        connect( dimensionBar, SIGNAL( foldingDimensionsChanged( std::vector<std::vector<int> >) ), data, SLOT( foldingDimensionsChanged( std::vector<std::vector<int> >) ) );
        connect( dimensionBar, SIGNAL( splitLengthChanged( int ) ), data, SLOT( splitLengthChanged( int ) ) );
    }
    return dimensionBar;
}

void
SystemTopologyWidget::handleSelection( TreeWidgetItem* item )
{
    SystemTopologyToolBar::getInstance()->setRecentSelection( item->id );
}

/**
   marks the given item as selected in System tree
   @param add if true, adds the item to the previously selected items, otherwise deselects previous selections
 */
void
SystemTopologyWidget::selectItem( int systemID, bool add )
{
    TreeWidgetItem* newSelection;
    newSelection = systemTreeWidget->items[ systemID ];
    bool deselect = false;

    if ( add ) // handle deselection of one previously selected element
    {
        {      // check if selected element is child of a selected and collapsed parent
            TreeWidgetItem* parentCollapsed = 0;
            TreeWidgetItem* item            = newSelection->parent();
            while ( item != 0 ) // get uppermost selected and collapsed parent node
            {
                if ( !item->isExpanded() && item->isSelected() )
                {
                    parentCollapsed = item;
                }
                item = item->parent();
            }
            if ( parentCollapsed != 0 )
            {
                // All children off a selected expanded parent item are implicitly selected.
                // If such a child gets deselected, the parent has to be expanded and
                // deselected, and it's children have to be selected.
                item = newSelection->parent();
                while ( item != parentCollapsed )
                {
                    item->setExpanded( true );

                    // select all siblings of the collapsed parents of the deselected element
                    std::vector<TreeWidgetItem*> siblings = item->parent()->getChildren();
                    for ( uint i = 0; i < siblings.size(); ++i )
                    {
                        if ( siblings[ i ] != item )
                        {
                            systemTreeWidget->select( siblings[ i ] );
                            siblings[ i ]->setExpanded( false ); // implicitly select all children
                        }
                    }
                    item = item->parent();
                }
                item->setExpanded( true );

                systemTreeWidget->deselect( parentCollapsed );

                std::vector<TreeWidgetItem*> siblings = newSelection->parent()->getChildren();
                for ( uint i = 0; i < siblings.size(); ++i )
                {
                    if ( siblings[ i ] != newSelection )
                    {
                        systemTreeWidget->select( siblings[ i ] );
                    }
                }
                systemTreeWidget->updateSizes();
                deselect = true;
            }
        }

        // deselect previously selected item
        const std::vector<TreeWidgetItem*>& selectionVec = systemTreeWidget->selectedItems();
        for ( uint i = 0; i < selectionVec.size(); ++i )
        {
            if ( newSelection == selectionVec.at( i ) )
            {
                systemTreeWidget->deselect( newSelection );
                systemTreeWidget->updateSizes();
                deselect = true;
            }
        }
    }
    else // deselect all previously selected items
    {
        std::vector<TreeWidgetItem*> selectionVec = systemTreeWidget->selectedItems();
        for ( uint i = 0; i < selectionVec.size(); ++i )
        {
            systemTreeWidget->deselect( selectionVec.at( i ) );
        }
    }

    if ( !deselect )
    {
        std::string selection = newSelection->getText().toStdString();
        selection.erase( 0, selection.find( " " ) + 1 ); // remove prepended value
        emit systemTreeWidget->setMessage( QString( "Selected \"" ).append( QString::fromStdString( selection ) ).append( "\"" ) );

        if ( newSelection->isHidden() )
        {
            newSelection->setHidden( false );
        }
        TreeWidgetItem* item = newSelection->parent();
        while ( item != NULL )
        {
            if ( item->isHidden() )
            {
                item->setHidden( false );
            }
            if ( !item->isExpanded() )
            {
                item->setExpanded( true );
            }
            item = item->parent();
        }
        systemTreeWidget->select( newSelection );
        systemTreeWidget->updateSizes();
    }
    data->updateSelection();
    updateValueWidget();
}

/** scroll the contents of the Scrollplane at position (x,y) to be placed
 * at current mousePosition
 * @param x,y position to scroll to, ignored if negative
 */
void
SystemTopologyWidget::scrollTo( int x, int y )
{
    QPoint mouse = this->mapFromGlobal( QCursor::pos() );

    // tab has changed => don't use mouse position (mouse is on tab control) but center position
    if ( mouse.y() < 0 )
    {
        mouse.setX( this->size().width() / 2 );
        mouse.setY( this->size().height() / 2 );
    }

    if ( x > -1 )
    {
        QScrollBar* sb = horizontalScrollBar();
        sb->setValue( x - mouse.x() );
    }
    if ( y > -1 )
    {
        QScrollBar* sb = verticalScrollBar();
        // sb->setValue() scrolls val to uppermost
        sb->setValue( y - mouse.y() );
    }
}

//-----------------------------------------------------------------------

//updates the value widget below the tab widget ; todo: move to controller class
void
SystemTopologyWidget::updateValueWidget()
{
    ValueWidget* valueWidget = systemTreeWidget->getTabWidget()->getValueWidget();

    //in case the topology is empty or if no items are selected, just
    //clear the vlaue widget
    if ( data->getDim( 0 ) == 0 || data->getDim( 1 ) == 0 || data->getDim( 2 ) == 0 || systemTreeWidget->selectedItems().size() == 0 )
    {
        valueWidget->clear();
    }
    else
    {
        double _sigma    = 0.0;
        double _variance = 0.0;
        double _mean     = 0.0;

        double  tmp = 0.;
        QString trap_value;
        bool    _ok = true;

        unsigned N = 0;
        for ( int i = 0; i < ( int )systemTreeWidget->items.size(); i++ )
        {
            TreeWidgetItem* item = systemTreeWidget->items.at( i );
            if ( item->childCount() != 0 )
            {
                continue;
            }

            systemTreeWidget->getValue( item, item->isExpanded(), tmp, trap_value, _ok, ABSOLUTE );

            _mean += tmp;
            N++;
        }

        _mean = ( N != 0 ) ? _mean / N : 0.;

        for ( int i = 0; i < ( int )systemTreeWidget->items.size(); i++ )
        {
            TreeWidgetItem* item = systemTreeWidget->items.at( i );
            if ( item->childCount() != 0 )
            {
                continue;
            }

            systemTreeWidget->getValue( item, item->isExpanded(), tmp, trap_value,  _ok, ABSOLUTE );

            _variance += ( _mean - tmp ) * ( _mean - tmp );
        }
        _sigma = ( N != 0 ) ? sqrt( _variance / N ) : 0.;


        double value         = 0.0;
        double absoluteValue = 0.0;
        bool   ok            = false;
        bool   intType       = systemTreeWidget->hasIntegerType();

        for ( int i = 0; i < ( int )systemTreeWidget->selectedItems().size(); i++ )
        {
            TreeWidgetItem* item = systemTreeWidget->selectedItems().at( i );
            systemTreeWidget->getValue( item, false, tmp, trap_value, ok, ABSOLUTE );
            if ( ok )
            {
                absoluteValue += tmp;
            }

            if ( ok && systemTreeWidget->getTabWidget()->getValueModus() != ABSOLUTE )
            {
                systemTreeWidget->getValue( item, false, tmp, ok );
                if ( ok )
                {
                    value += tmp;
                }
            }
        }

        //compute value for minimal color
        double min;
        if ( data->getUserDefinedMinMaxValues() )
        {
            min = data->getUserMinValue();
        }
        else if ( systemTreeWidget->tabWidget->getValueModus() == ABSOLUTE )
        {
            min = data->getMinValue();
        }
        else
        {
            min = 0.0;
        }

        //compute value for minimal color
        double max;
        if ( data->getUserDefinedMinMaxValues() )
        {
            max = data->getUserMaxValue();
        }
        else if ( systemTreeWidget->tabWidget->getValueModus() == ABSOLUTE )
        {
            max = data->getMaxValue();
        }
        else
        {
            max = 100.0;
        }

        if ( !ok )
        {
            //if one of the values was undefined, display only the min and
            //max values
            valueWidget->update( min, max, intType, false, _mean, _sigma );
        }
        else if ( systemTreeWidget->getTabWidget()->getValueModus() == ABSOLUTE )
        {
            //in the absolute value mode display the absolute value (with
            //the min/max values)
            valueWidget->update( min, max, absoluteValue, intType, false, _mean, _sigma );
        }
        else
        {
            //in value modes other that the absolute value mode display the
            //current minvalue/value/maxvalue plus absolute
            //minvalue/value/maxvalue
            valueWidget->update( 0.0, 100.0, value, data->getMinAbsValue(), data->getMaxAbsValue(), absoluteValue, intType, false, _mean, _sigma );
        }
    }
}


/****************** slots for menuitems / TabWidget ***************************/

void
SystemTopologyWidget::rescale()
{
    transform->rescale( false );
}

//set the line type for topologies: black, gray, white, or no lines
void
SystemTopologyWidget::setLineType( LineType lineType )
{
    data->setLineType( lineType );
}

//toggle the state if unused planes should be displayed or not
void
SystemTopologyWidget::toggleUnused()
{
    data->toggleUnused();
}

void
SystemTopologyWidget::valuesChanged()
{
    data->updateColors();
    view->updateDrawing();
    updateValueWidget();
}

void
SystemTopologyWidget::updateColors()
{
    data->updateColors();
    view->updateDrawing();
    updateValueWidget();
}

/************************ settings ***************************/

//save topology settings
//
void
SystemTopologyWidget::saveSettings( QSettings& settings )
{
    transform->saveSettings( settings, topologyId );
    dimensionBar->saveSettings( settings, topologyId );
}

bool
SystemTopologyWidget::loadSettings( QSettings& settings )
{
    dimensionBar->loadSettings( settings, topologyId );
    transform->loadSettings( settings, topologyId );
    return true;
}

void
SystemTopologyWidget::setActive()
{
    SystemTopologyToolBar::getInstance()->setParent( this );
}
