/****************************************************************************
**  CUBE        http://www.score-p.org/                                    **
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2014                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/



#include <cstring>
#include <iostream>
#include <cassert>
#include <cmath>
// #ifndef CUBE_COMPRESSED
#include <fstream>
// #else
#include "CubeZfstream.h"
// #endif

#include <QMouseEvent>
#include <QString>
#include <QHeaderView>
#include <QInputDialog>
#include <QMessageBox>
#include <QFileDialog>
#include <QTextStream>
#include <QHBoxLayout>
#include <QPushButton>
#include <QApplication>
#include <QSettings>
#include <QScrollBar>
#include <QtNetwork>
#include <QDialogButtonBox>
#include <QPainter>
#include <QRadioButton>
#include <QFontDatabase>
#include <QRegExp>
#include <QTextEdit>
#include <QClipboard>
#include <QFontDialog>
#include <QFontMetrics>


#include "SystemTreeWidget.h"
#include "CallTreeWidget.h"
#include "TreeWidgetItem.h"
#include "Constants.h"
#include "TypedWidget.h"
#include "TabWidget.h"
#include "PrecisionWidget.h"
#include "ValueWidget.h"
#include "HelpBrowser.h"
#include "ScrollArea.h"
#include "MessageWidget.h"

#ifdef WITH_EXPERIMENTAL_VIEWS
#  include "2DPlotWidget.h"
#  include "ColorMapPlotWidget.h"
#endif   /* WITH_EXPERIMENTAL_VIEWS */

#include "Cube.h"
#include "CubeMetric.h"
#include "CubeCnode.h"
#include "CubeRegion.h"
#include "CubeMachine.h"
#include "CubeNode.h"
#include "CubeProcess.h"
#include "CubeThread.h"
#include "CubeVertex.h"
#include "CubeError.h"
#include "CubeValues.h"

using namespace std;
using namespace cube;



//special case:
//compute the current value for a Systemtree widget item
//if the systen tree is in the center
//
SystemTreeWidget::SystemTreeWidget( QWidget*            parent,
                                    QString             cubeFileName,
                                    cubeparser::Driver* driver,
                                    Statistics*         statistics ) : TreeWidget( parent, SYSTEMTREE, cubeFileName, driver, statistics )
{
    subsetCombo = NULL;
    container   = new SplitterContainer();
    container->addWidget( scrollArea );

    // set lower splitter element to minimum size (1 pixel => replaced by minimumSize())
    QList<int> sizeList;
    sizeList << container->size().height() << 1;
    container->setSizes( sizeList );
}

void
SystemTreeWidget::computeValue( list_of_metrics& mv,  TreeWidgetItem* item )
{
    //         if (!item->isInvalid()) return;
    if ( !item->isInvalid() && !item->isExpanded() )
    {
        return;
    }


    if ( item->isInvalid() )
    {
        assert(
            item->type == SYSTEMTREENODEITEM ||
            item->type == LOCATIONGROUPITEM ||
            item->type == LOCATIONITEM
            );


        cube::Sysres* sysres = ( ( cube::Sysres* )( item->cubeObject ) );

        if ( item->totalValue_adv != NULL )
        {
            item->totalValue_adv->Free();
        }
        if ( item->ownValue_adv != NULL )
        {
            item->ownValue_adv->Free();
        }
        item->totalValue_adv = cube->get_sev_adv( mv, sysres, cube::CUBE_CALCULATE_INCLUSIVE );
        item->ownValue_adv   = cube->get_sev_adv( mv, sysres, cube::CUBE_CALCULATE_EXCLUSIVE );
        item->totalValue     = ( item->totalValue_adv != NULL ) ? item->totalValue_adv->getDouble() : 0.;
        item->ownValue       = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->getDouble() : 0.;
        item->valueCollapsed = item->totalValue;
        item->valueExpanded  = item->ownValue;

        item->singleValue = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->singleValue() : true;

        if ( item->parent() != NULL )
        {
            item->rootValue = item->parent()->rootValue;
        }
        else
        {
            item->rootValue = item->totalValue;
        }
    }

    for ( int i = 0; i < ( int )( item->childCount() ); i++ )
    {
        computeValue( mv, item->child( i ) );
    }
}












//special case:
//compute the current value for a Systemtree widget item
//if the systen tree is the last panel
//
void
SystemTreeWidget::computeValue( list_of_metrics& mv, list_of_cnodes& cv,  TreeWidgetItem* item )
{
    //         if (!item->isInvalid()) return;
    if ( !item->isInvalid() && !item->isExpanded() )
    {
        return;
    }


    if ( item->isInvalid() )
    {
        assert(
            item->type == SYSTEMTREENODEITEM ||
            item->type == LOCATIONGROUPITEM ||
            item->type == LOCATIONITEM
            );

        cube::Sysres* sysres = ( ( cube::Sysres* )( item->cubeObject ) );

        if ( item->totalValue_adv != NULL )
        {
            item->totalValue_adv->Free();
        }
        if ( item->ownValue_adv != NULL )
        {
            item->ownValue_adv->Free();
        }
        item->totalValue_adv = cube->get_sev_adv( mv, cv, sysres, cube::CUBE_CALCULATE_INCLUSIVE );
        item->ownValue_adv   = cube->get_sev_adv( mv, cv, sysres, cube::CUBE_CALCULATE_EXCLUSIVE );
        item->totalValue     = ( item->totalValue_adv != NULL ) ? item->totalValue_adv->getDouble() : 0.;
        item->ownValue       = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->getDouble() : 0.;
        item->valueCollapsed = item->totalValue;
        item->valueExpanded  = item->ownValue;
        item->singleValue    = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->singleValue() : true;
    }

    if ( item->parent() != NULL )
    {
        item->rootValue = item->parent()->rootValue;
    }
    else
    {
        item->rootValue = item->totalValue;
    }

    for ( int i = 0; i < ( int )( item->childCount() ); i++ )
    {
        computeValue( mv, cv, item->child( i ) );
    }
}








//special case:
//compute the current value for a Systemtree widget item
//if the systen tree is in the center
//
void
SystemTreeWidget::computeValue( list_of_metrics& mv, list_of_regions& rv,  TreeWidgetItem* item )
{
    //         if (!item->isInvalid()) return;
    if ( !item->isInvalid() && !item->isExpanded() )
    {
        return;
    }


    if ( item->isInvalid() )
    {
        assert(
            item->type == SYSTEMTREENODEITEM ||
            item->type == LOCATIONGROUPITEM ||
            item->type == LOCATIONITEM
            );



        cube::Sysres* sysres = ( ( cube::Sysres* )( item->cubeObject ) );

        if ( item->totalValue_adv != NULL )
        {
            item->totalValue_adv->Free();
        }
        if ( item->ownValue_adv != NULL )
        {
            item->ownValue_adv->Free();
        }
        item->totalValue_adv = cube->get_sev_adv( mv, rv, sysres, cube::CUBE_CALCULATE_INCLUSIVE );
        item->ownValue_adv   = cube->get_sev_adv( mv, rv, sysres, cube::CUBE_CALCULATE_EXCLUSIVE );
        item->totalValue     = ( item->totalValue_adv != NULL ) ? item->totalValue_adv->getDouble() : 0.;
        item->ownValue       = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->getDouble() : 0.;
        item->valueCollapsed = item->totalValue;
        item->valueExpanded  = item->ownValue;
        item->singleValue    = ( item->ownValue_adv != NULL ) ? item->ownValue_adv->singleValue() : true;
    }
    if ( item->parent() != NULL )
    {
        item->rootValue = item->parent()->rootValue;
    }
    else
    {
        item->rootValue = item->totalValue;
    }

    for ( int i = 0; i < ( int )( item->childCount() ); i++ )
    {
        computeValue( mv, rv, item->child( i ) );
    }
}








//compute all absolute values for this tree's items
//
void
SystemTreeWidget::computeBasicValues()
{
//                 cout << "system SDFFSDG " << endl;
    list_of_regions      rv;
    list_of_cnodes       cv;
    list_of_sysresources sv;
    ;
    list_of_metrics mv;



    bool unset_regions = false; ///  marks, that some regions got status "subroutines" and  have to be reset after calculation;
    //set all values of all items to 0.0
    initializeValues();

//         cout << "Calculate bsic values for System ! " << endl;
    int index, metricIndex, callIndex, systemIndex;
    getIndices( index, metricIndex, callIndex, systemIndex );


    if ( systemIndex == 0 )
    {
        return;                   // situation, systemTree, XXX, XXX -> Systemtree doesn't show anything
    }
    if ( systemIndex == 1 && metricIndex == 2 )
    {
        return;                                          // situation, Calltree, System, Metric -> systemtree doesnt show anything
    }
    if ( tabWidget->getMetricTree()->Items().size() == 0 ||
         tabWidget->getCallTree()->Items().size() == 0 ||
         tabWidget->getSystemTree()->Items().size() == 0
         )
    {
        return;
    }
    if ( tabWidget->getMetricTree()->selectedItems().size() == 0 ||
         tabWidget->getCallTree()->selectedItems().size() == 0 ||
         tabWidget->getSystemTree()->selectedItems().size() == 0
         )
    {
        return;
    }

    if ( systemIndex == 1 ) // situation  Metric, System, Calltree -> standard
    {
        std::vector<TreeWidgetItem*> sel_metrics = tabWidget->getMetricTree()->selectedItems();
        for ( std::vector<TreeWidgetItem*>::iterator miter = sel_metrics.begin(); miter != sel_metrics.end(); miter++ )
        {
            metric_pair mp;
            mp.first  = ( cube::Metric* )( ( *miter )->cubeObject );
            mp.second = ( ( *miter )->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
            mv.push_back( mp );
        }

        for ( int i = 0; i < ( int )top.size(); i++ )
        {
            computeValue( mv, top[ i ] );
        }
    }


    if ( systemIndex == 2 ) // situation  Metric, Calltree, System  or Calltree, Metrictree, System
    {
        std::vector<TreeWidgetItem*> sel_metrics = tabWidget->getMetricTree()->selectedItems();
        for ( std::vector<TreeWidgetItem*>::iterator miter = sel_metrics.begin(); miter != sel_metrics.end(); miter++ )
        {
            metric_pair mp;
            mp.first  = ( cube::Metric* )( ( *miter )->cubeObject );
            mp.second = ( ( *miter )->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
            mv.push_back( mp );
        }
        if ( tabWidget->getCallTabWidget()->getCurrentCallTree()->getType() == CALLFLAT )
        {
            std::vector<TreeWidgetItem*> sel_regions = tabWidget->getCurrentCallTree()->selectedItems();
            for ( std::vector<TreeWidgetItem*>::iterator riter = sel_regions.begin(); riter != sel_regions.end(); riter++ )
            {
                region_pair rp;
                rp.first =   ( cube::Region* )( ( *riter )->cubeObject );
                if ( rp.first == NULL )
                {
                    unset_regions = true;
                    cube::Region* reg = ( cube::Region* )( ( *riter )->parent()->cubeObject );
                    reg->set_as_subroutines();
                    rp.first = reg;
                }
                rp.second = ( ( *riter )->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
                rv.push_back( rp );
            }
            for ( int i = 0; i < ( int )top.size(); i++ )
            {
                computeValue( mv, rv, top[ i ] );
            }
        }
        else
        {
            CallTreeWidget*              callTree      = dynamic_cast<CallTreeWidget*> ( tabWidget->getCallTree() );
            std::vector<TreeWidgetItem*> sel_callnodes = callTree->selectedItemsAndIterations();
            for ( std::vector<TreeWidgetItem*>::iterator citer = sel_callnodes.begin(); citer != sel_callnodes.end(); citer++ )
            {
                cnode_pair cp;
                cp.first  = ( cube::Cnode* )( ( *citer )->cubeObject );
                cp.second = ( ( *citer )->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
                cv.push_back( cp );
            }
            for ( int i = 0; i < ( int )top.size(); i++ )
            {
                computeValue( mv, cv,  top[ i ] );
            }
        }
    }


    if ( unset_regions )
    {
        for ( list_of_regions::iterator riter = rv.begin(); riter != rv.end(); riter++ )
        {
            ( *riter ).first->unset_as_subroutines();
        }
    }


    for ( int i = 0; i < ( int )top.size(); i++ )
    {
        validateValue( top[ i ] );
    }



    if ( type == SYSTEMTREE )
    {
//         double maxMachine = 0.0;
//         double minMachine = 0.0;
//         double maxNode    = 0.0;
//         double minNode    = 0.0;
//         double maxProcess = 0.0;
//         double minProcess = 0.0;
//         double maxThread  = 0.0;
//         double minThread  = 0.0;
        double tmp = 0.0;

//         bool   machineSeen = false;
//         bool   nodeSeen    = false;
//         bool   processSeen = false;
//         bool   threadSeen  = false;
        vector<double> min_vector;
        vector<double> max_vector;
        vector<bool>   depth_seen;
        for ( unsigned i = 0; i < items.size(); i++ )
        {
            size_t depth = items[ i ]->getDepth();

            if ( min_vector.size() <= depth )
            {
                min_vector.resize( depth + 1 );
            }
            if ( max_vector.size() <= depth )
            {
                max_vector.resize( depth + 1 );
            }
            while ( depth_seen.size() <= depth )
            {
                depth_seen.push_back( false );
            }


            tmp = items[ i ]->totalValue;

            if ( !depth_seen[ depth ] )
            {
                depth_seen[ depth ] = true;
                min_vector[ depth ] = tmp;
                max_vector[ depth ] = tmp;
            }
            else
            {
                min_vector[ depth ] = ( min_vector[ depth ] > tmp ) ? tmp : min_vector[ depth ];
                max_vector[ depth ] = ( max_vector[ depth ] < tmp ) ? tmp : max_vector[ depth ];
            }
        }

        for ( unsigned i = 0; i < items.size(); i++ )
        {
            int depth = items[ i ]->getDepth();
            items[ i ]->minValue = min_vector[ depth ];
            items[ i ]->maxValue = max_vector[ depth ];
        }
    }
}
//end of computeBasicValues()


/**
 * @brief compute the value of one element
 */
double
SystemTreeWidget::computeValue( TreeWidgetItem* metricItem, cube::CalculationFlavour metricState,
                                TreeWidgetItem* callItem, cube::CalculationFlavour callState,
                                TreeWidgetItem* systemItem, cube::CalculationFlavour systemState )
{
    double value = 0.;

    assert( systemItem->type == SYSTEMTREENODEITEM ||
            systemItem->type == LOCATIONGROUPITEM ||
            systemItem->type == LOCATIONITEM
            );

    list_of_regions      rv;
    list_of_cnodes       cv;
    list_of_sysresources sv;
    list_of_metrics      mv;

    int index, metricIndex, callIndex, systemIndex;
    getIndices( index, metricIndex, callIndex, systemIndex );

    if ( systemIndex == 0 )
    {
        return 0;                // situation, systemTree, XXX, XXX -> Systemtree doesn't show anything
    }
    if ( systemIndex == 1 && metricIndex == 2 )
    {
        return 0;                // situation, Calltree, System, Metric -> systemtree doesnt show anything
    }
    if ( tabWidget->getMetricTree()->Items().size() == 0 ||
         tabWidget->getCallTree()->Items().size() == 0 ||
         tabWidget->getSystemTree()->Items().size() == 0
         )
    {
        return 0;
    }
    if ( tabWidget->getMetricTree()->selectedItems().size() == 0 ||
         tabWidget->getCallTree()->selectedItems().size() == 0 ||
         tabWidget->getSystemTree()->selectedItems().size() == 0
         )
    {
        return 0;
    }

    // set metric value data
    metric_pair mp;
    mp.first  = ( cube::Metric* )metricItem->cubeObject;
    mp.second = metricState;
    mv.push_back( mp );

    if ( systemIndex == 1 ) // situation  Metric, System, Calltree -> standard
    {
        value = cube->get_sev( mv, ( cube::Sysres* )systemItem->cubeObject, systemState );
    }
    else if ( systemIndex == 2 ) // situation  Metric, Calltree, System  or Calltree, Metrictree, System
    {
        if ( tabWidget->getCallTabWidget()->getCurrentCallTree()->getType() == CALLFLAT )
        {
            region_pair rp;
            rp.first  =   ( cube::Region* )callItem->cubeObject;
            rp.second = callState;
            rv.push_back( rp );

            value = cube->get_sev( mv, rv, ( cube::Sysres* )systemItem->cubeObject, systemState );
        }
        else // CALLTREE
        {
            cnode_pair cp;
            cp.first  = ( cube::Cnode* )callItem->cubeObject;
            cp.second = callState;
            cv.push_back( cp );

            value = cube->get_sev( mv, cv, ( cube::Sysres* )systemItem->cubeObject, systemState );
        }
    }
    return value;
}

/**
 * @brief TreeWidget::getVisitedSubset
 * @return predefined subset: all thread with visited > 0
 */
std::vector<TreeWidgetItem*>
SystemTreeWidget::getVisitedSubset()
{
    std::vector<TreeWidgetItem*> visited;
    double                       val      = 0;
    TreeWidget*                  callTree = tabWidget->getCurrentCallTree();
    TreeWidgetItem*              visit    =  tabWidget->getMetricTree()->findItems( "Visit" ).at( 0 );

    for ( unsigned i = 0; i < this->items.size(); ++i )
    {
        if ( ( this->items[ i ]->type == LOCATIONGROUPITEM && this->items[ i ]->childCount() == 0 )
             || ( this->items[ i ]->type == LOCATIONITEM ) )
        {
            for ( unsigned j = 0; j < callTree->selectedItems().size(); ++j )
            {
                TreeWidgetItem* callItem = callTree->selectedItems().at( j );

                cube::CalculationFlavour callTreeState;
                if ( !callItem->isExpanded() )
                {
                    callTreeState =  cube::CUBE_CALCULATE_INCLUSIVE;
                }
                else
                {
                    callTreeState = CUBE_CALCULATE_EXCLUSIVE;
                }

                val = this->computeValue( visit, cube::CUBE_CALCULATE_INCLUSIVE,
                                          callItem, callTreeState,
                                          this->items[ i ], cube::CUBE_CALCULATE_INCLUSIVE );
                if ( val > 0 )
                {
                    break;
                }
            }
            if ( val > 0 )
            {
                visited.push_back( this->items[ i ] );
            }
        }
    }
    return visited;
}

QIcon
SystemTreeWidget::icon() const
{
    return QIcon( ":images/tree-icon.png" );
}

QWidget*
SystemTreeWidget::widget()
{
    return container;
}

void
SystemTreeWidget::updateSelection()
{
    vector<TreeWidgetItem*> subsetItems = getActiveSubset();
    subsetItemCount.insert( getSubsetName(), subsetItems.size() );

    if ( subsetItems.size() == items.size() ) // dont't mark everything
    {
        subsetItems.clear();
    }

    markTreeItems( subsetItems );
}

void
SystemTreeWidget::setSubsetCombo( QComboBox* combo )
{
    subsetCombo = combo;

    fillSubsetCombo();
    resetSubsetCombo();

    connect( subsetCombo, SIGNAL( currentIndexChanged( int ) ), this, SLOT( updateSubsetCombo() ) );
    connect( this, SIGNAL( selectionChanged() ), this, SLOT( resetSubsetCombo() ) );
    connect( this, SIGNAL( definedSubsetsChanged( const QString & ) ),
             this, SLOT( fillSubsetCombo( const QString & ) ) );
}

void
SystemTreeWidget::setActive()
{
    if ( !subsetCombo )
    {
        return;
    }
    QSplitter* oldParent = ( QSplitter* )subsetCombo->parentWidget();
    QSplitter* newParent = container;
    QList<int> sizeList  = oldParent->sizes();

    // share the same instance of QCombobox between BoxPlot and SystemTree -> change the parent widget
    subsetCombo->setParent( newParent );

    // save collapsed state
    newParent->setSizes( sizeList );
}

/**
 * @brief TreeWidget::getActiveSubset
 * @return the TreeWidgetItems which belong the the currently selected subset
 */
std::vector<TreeWidgetItem*>
SystemTreeWidget::getActiveSubset()
{
    QString name = getSubsetName();
    if ( name == "All" )
    {
        return items;
    }
    else if ( name == "Visited" )
    {
        return getVisitedSubset();
    }
    else
    {
        return definedSubsets.value( name );
    }
}

int
SystemTreeWidget::getSubsetItemCount( const QString& subsetName )
{
    int count = 0;
    if ( subsetName == "Visited" )
    {
        vector<TreeWidgetItem*> subsetItems = getVisitedSubset(); // todo: avoid unnecessary calls
        count = subsetItems.size();
    }
    else if ( subsetName == "All" )
    {
        for ( unsigned i = 0; i < items.size(); ++i )
        {
            if ( ( items[ i ]->type == LOCATIONGROUPITEM && items[ i ]->childCount() == 0 )
                 || ( items[ i ]->type == LOCATIONITEM ) )
            {
                count++;
            }
        }
    }
    return count;
}

/**
 * @brief removes additional description from subset name
 * @param subsetName if empty, the currently selected subset is used
 */
QString
SystemTreeWidget::getSubsetName( const QString& subsetName )
{
    if ( !subsetCombo )
    {
        return QString();
    }

    QString name = subsetName;
    if ( name.isEmpty() )   // get name of selected Element
    {
        name = subsetCombo->currentText();
    }
    name.remove( QRegExp( "\\s*\\([0-9]* elements\\)$" ) );
    return name;
}

void
SystemTreeWidget::resetSubsetCombo()
{
    subsetCombo->setCurrentIndex( 0 );
}

/**
 * @brief SystemTreeWidget::fillSubsetCombo
 * @param name if not empty, the entry with the given name selected
 */
void
SystemTreeWidget::fillSubsetCombo( const QString& name )
{
    if ( !subsetCombo )
    {
        return;
    }
    subsetCombo->setModel( &subsetModel );

    QStringList list;
    list << "All";
    list << "Visited";

    int idx = -1;
    if ( tabWidget->getSystemTree() != 0 )
    {
        const QHash<QString, std::vector<TreeWidgetItem*> >& definedSubsets = tabWidget->getSystemTree()->getDefinedSubsets();
        QStringList                                          keys           = definedSubsets.keys();
        keys.sort();
        for ( int i = 0; i < keys.size(); ++i )
        {
            uint size = definedSubsets.value( keys.at( i ) ).size();
            list << keys.at( i ) + " (" + QString::number( size ) + " elements)";
            if ( !name.isEmpty() && keys.at( i ) == name )   // select new item
            {
                idx = list.size() - 1;
            }
        }
    }
    subsetModel.setStringList( list );
    if ( idx >= 0 )
    {
        subsetCombo->setCurrentIndex( idx );
    }
}

/**
   updates the number of processes/threads
   @oaram updateAll update even if currently not selected
 */
void
SystemTreeWidget::updateSubsetCombo()
{
    if ( subsetCombo && tabWidget->getSystemTree() )
    {
        // update "All" (only initalisation necessary)
        QString name = "All";
        int     idx  = subsetCombo->findText( name, Qt::MatchExactly );
        if ( idx >= 0 )
        {
            int     count = getSubsetItemCount( name );
            QString txt   = name + " (" + QString::number( count ) + " elements)";
            if ( idx >= 0 )
            {
                subsetModel.setData( subsetModel.index( idx ), txt );
            }
        }

        // update "Visited"
        name = "Visited";
        idx  = subsetCombo->findText( name, Qt::MatchContains );
        if ( idx == subsetCombo->currentIndex() )
        {   //visible
            int     visited = getSubsetItemCount( name );
            QString txt     = name + " (" + QString::number( visited ) + " elements)";
            if ( idx >= 0 )
            {
                subsetModel.setData( subsetModel.index( idx ), txt );
            }
        }
        else
        {
            subsetModel.setData( subsetModel.index( idx ), name );
        }
    }
}


void
SystemTreeWidget::valuesChanged()
{
    TreeWidget::valuesChanged();
    updateSubsetCombo();
}
