/****************************************************************************
**  CUBE        http://www.score-p.org/                                    **
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2014                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/



#include <QScrollBar>
#include <QLabel>

#include "TabWidget.h"
#include "MetricTreeWidget.h"
#include "CallTreeWidget.h"
#include "FlatTreeWidget.h"
#include "SystemTreeWidget.h"

#include "SystemTopologyWidget.h"
#include "SystemTopologyToolBar.h"
#include "PrecisionWidget.h"
#include "Statistics.h"
#include "SystemBoxPlot.h"
#include "CubeMetric.h"
#include "CubeCartesian.h"

#include <cassert>
#include <qdebug.h>

TabWidget::TabWidget( TabWidgetType    type,
                      ColorWidget*     colorWidget,
                      ValueWidget*     valueWidget,
                      PrecisionWidget* precisionWidget ) : QTabWidget()
{
    //general initializations
    valueModus      = ABSOLUTE;
    cube            = NULL;
    metricTabWidget = NULL;
    callTabWidget   = NULL;
    systemTabWidget = NULL;

    this->type            = type;
    this->colorWidget     = colorWidget;
    this->valueWidget     = valueWidget;
    this->precisionWidget = precisionWidget;

    //tree-specific initializations
    treeFont = font();
    spacing  = 2;

    //topology-specific initializations
    currentSysTopWidget      = NULL;
    whiteForZero             = true;
    lineType                 = BLACK_LINES;
    antialiasing             = false;
    showUnusedTopologyPlanes = true;
    systemBoxPlot            = NULL;
    systemTreeIndex          = -1;
    systemBoxPlotIndex       = -1;
#ifdef _AIX
    selectionSyntax = FRAME_SELECTION;
#else
    selectionSyntax = BACKGROUND_SELECTION;
#endif
    metricTree = 0, callTree = 0, flatTree = 0, systemTree = 0;
}

TabWidget::~TabWidget()
{
}

int
TabWidget::addTab( QWidget* widget, TabInterface* tc )
{
    tc->initialize( cube );
    tc->setLineType( lineType );
    tabList.append( tc );
    return QTabWidget::addTab( widget, tc->icon(), tc->label() );
}

int
TabWidget::addTreeTab( TreeWidget* treeWidget )
{
    treeWidget->setFont( treeFont );
    treeWidget->setSpacing( spacing );
    treeWidget->setTabWidget( this );

    connect( treeWidget, SIGNAL( setMessage( QString ) ), this, SIGNAL( setMessage( QString ) ) );

    return addTab( treeWidget->widget(), treeWidget );
}

TabInterface*
TabWidget::tabInterface( int idx ) const
{
    return tabList.at( idx );
}

/**************** get methods ******************/

//return the current value modus, see constants.h for the ValueModus type
//
ValueModus
TabWidget::getValueModus()
{
    return valueModus;
}

//get different widgets

ColorWidget*
TabWidget::getColorWidget()
{
    return colorWidget;
}
PrecisionWidget*
TabWidget::getPrecisionWidget()
{
    return precisionWidget;
}
ValueWidget*
TabWidget::getValueWidget()
{
    return valueWidget;
}

/**
   returns the position of the current TabWidget in the parent QSplitter widget (0-2)
 */
int
TabWidget::getOrder()
{
    QWidget* parent = this->parentWidget();
    QWidget* child  = NULL;
    while ( parent != NULL && !dynamic_cast<QSplitter*> ( parent ) )
    {
        child  = parent;
        parent = parent->parentWidget();
    }

    int idx = -1;
    if ( parent != NULL )
    {
        QSplitter* splitter = dynamic_cast<QSplitter*> ( parent );
        idx = splitter->indexOf( child );
    }
    return idx;
}

TreeWidget*
TabWidget::getMetricTree()
{
    return this->metricTabWidget->metricTree;
}

TreeWidget*
TabWidget::getCallTree()
{
    return this->callTabWidget->callTree;
}

TreeWidget*
TabWidget::getSystemTree()
{
    return this->systemTabWidget->systemTree;
}

//return the current tab element of the call tab (call tree or flat call profile)
TreeWidget*
TabWidget::getCurrentCallTree()
{
    int index = callTabWidget->currentIndex();
    return ( index == 0 ) ? this->callTabWidget->callTree : this->callTabWidget->flatTree;
}

//return the ith external metric item
//
TreeWidgetItem*
TabWidget::getExternalMetricItem( int i )
{
    return externalMetricItems[ i ];
}

//return the name if the ith external metric item
//
std::string
TabWidget::getExternalMetricName( int i )
{
    return externalMetricNames[ i ];
}

//get the different tabs

TabWidget*
TabWidget::getMetricTabWidget()
{
    return metricTabWidget;
}
TabWidget*
TabWidget::getCallTabWidget()
{
    return callTabWidget;
}
TabWidget*
TabWidget::getSystemTabWidget()
{
    return systemTabWidget;
}


//return if unused topology planes should be displayed
//
bool
TabWidget::getShowUnusedTopologyPlanes()
{
    return showUnusedTopologyPlanes;
}

//return if lines in topology display should be antialiased
//
bool
TabWidget::getAntialiasing()
{
    return antialiasing;
}

//return the line style for topologies
//(black, gray, white or no lines, see constants.h for the LineType type)
//
LineType
TabWidget::getLineType()
{
    return lineType;
}

//tell how zero values in topologies are colored
//(white or minimal color)
//
bool
TabWidget::getWhiteForZero()
{
    return whiteForZero;
}

//return the current method for marking selected tree items
//(by blue background or by framing, see constants.h for the SelectionSyntax type)
//
SelectionSyntax
TabWidget::getSelectionSyntax()
{
    return selectionSyntax;
}


/******************* set methods ***************/

//set the value modus, see constants.h for the ValueModus type
//
void
TabWidget::setValueModus( ValueModus valueModus )
{
    this->valueModus = valueModus;

    if ( this->count() == 0 )
    {
        return;
    }

    tabInterface( currentIndex() )->valueModusChanged();
}

//set the different tab widgets;
//they are needed to determine the different trees

void
TabWidget::setMetricTabWidget( TabWidget* widget )
{
    metricTabWidget = widget;
}
void
TabWidget::setCallTabWidget( TabWidget* widget )
{
    callTabWidget = widget;
}
void
TabWidget::setSystemTabWidget( TabWidget* widget )
{
    systemTabWidget = widget;
}

//set font and line spacing for all tree displays
void
TabWidget::setTreeDisplay( QFont font, int spacing )
{
    treeFont      = font;
    this->spacing = spacing;

    QList<TreeWidget*> trees = QList<TreeWidget*>() << metricTree << callTree << flatTree << systemTree;

    foreach( TreeWidget * treeWidget, trees )
    {
        if ( treeWidget != NULL )
        {
            treeWidget->setFont( treeFont );
            treeWidget->setSpacing( spacing );
            treeWidget->updateSizes();
        }
    }
}

//set how zero values in topologies should be colored
//(white or minimal color)
//
void
TabWidget::setWhiteForZero( bool whiteForZero )
{
    this->whiteForZero = whiteForZero;
}

//set how selected tree items should be marked
//(by blue background or by framing, see constants.h for the SelectionSyntax type)
//
void
TabWidget::setSelectionSyntax( SelectionSyntax selectionSyntax )
{
    this->selectionSyntax = selectionSyntax;
    displayItems();
}


/**************** initialization and clean-up ***************/

//initialize according to the type and the cube database
//
void
TabWidget::initialize( cube::Cube* cube, QString fileName,
                       cubeparser::Driver* driver, Statistics* statistics )
{
    tabList.clear();
    this->cube = cube;

    setUsesScrollButtons( true );
    setIconSize( QSize( 20, 20 ) );

    if ( type == METRICTAB )
    {
        metricTree = new MetricTreeWidget( this, fileName, driver, statistics );
        addTreeTab( metricTree );
    }
    else if ( type == CALLTAB )
    {
        //call tabs have a call tree and a flat call profile
        callTree = new CallTreeWidget( this, fileName, driver, statistics );
        addTreeTab( callTree );

        flatTree = new FlatTreeWidget( this );
        addTreeTab( flatTree );

        if ( !( cube->is_flat_tree_enabled() ) )
        {
            setTabEnabled( count() - 1, false );
            flatTree->setEnabled( false );
        }
    }
    else if ( type == SYSTEMTAB )
    {
        QComboBox* subsetCombo = new QComboBox();
        subsetCombo->setWhatsThis(
            tr( "The Boxplot uses the currently selected subset of threads when determining its statistics."
                " Other defined subsets can be chosen from the combobox menu, such as \"All\" threads or"
                " \"Visited\" threads for only threads that visited the currently selected callpath."
                " Additional subsets can be defined from the System Tree with the \"Define subset\" context menu"
                " using the currently selected threads via multiple selection (control + left mouseclick)"
                " or with the \"Find items\" context menu selection option." ) );

        //system tabs have a system tree and topologies
        {
            systemTree      = new SystemTreeWidget( this );
            systemTreeIndex = addTreeTab( systemTree );

            systemTree->setSubsetCombo( subsetCombo );
            connect( subsetCombo, SIGNAL( currentIndexChanged( int ) ), this, SLOT( displayItems() ) );

            subsetCombo->setParent( systemTree->widget() ); // first visible in systemTree
        }
        {                                                   // box plot tab
            systemBoxPlot      = new SystemBoxPlot( this, systemTree );
            systemBoxPlotIndex = addTab( systemBoxPlot->widget(), systemBoxPlot );
            systemBoxPlot->setSubsetCombo( subsetCombo );
        }
        {
            SystemTopologyWidget* systemTopologyWidget;
            unsigned              numTopologies = cube->get_cartv().size();
            if ( numTopologies > 0 )
            {
                SystemTopologyToolBar::getInstance()->show();
            }
            else
            {
                SystemTopologyToolBar::getInstance()->hide();
            }

            for ( unsigned i = 0; i < numTopologies; i++ )
            {
                systemTopologyWidget = new SystemTopologyWidget( systemTree, i );
                addTab( systemTopologyWidget->widget(), systemTopologyWidget );
            }
        } // SYSTEMTAB
    }
    connect( this, SIGNAL( currentChanged( int ) ), this, SLOT( onCurrentChanged( int ) ) );
}


//called when files are closed;
//clear the tab, no data is shown
//
void
TabWidget::cleanUp()
{
    currentSysTopWidget         = NULL;
    systemTabWidget->systemTree = NULL;
    metricTabWidget->metricTree = NULL;
    callTabWidget->callTree     = NULL;

    disconnect( this, SIGNAL( currentChanged( int ) ), this, SLOT( onCurrentChanged( int ) ) );

    //remove tab elements
    while ( count() > 0 )
    {
        QWidget* child = widget( 0 );
        removeTab( 0 );
        delete child;
    }
}


//clean up external values;
//called when external files are closed
//
void
TabWidget::clearExternalValues()
{
    for ( int i = 0; i < ( int )( externalMetricItems.size() ); i++ )
    {
        delete externalMetricItems[ i ];
    }
    externalMetricItems.clear();
    externalMetricNames.clear();
}


/************** value computations ******************/


//(re-)compute the values for the current tab element;
//in case of system tabs, compute the values for the system tree,
//since the topologies use the tree values
//
void
TabWidget::computeValues()
{
    if ( count() > 0 )
    {
        int index = ( type == SYSTEMTAB ? 0 : currentIndex() );

        tabInterface( index )->recalculateItems();

        displayItems();
    }
}

//compute the needed external values from the "cubeExternal"
//database for the external percentage value modus
//
void
TabWidget::computeExternalValues( cube::Cube* cubeExternal )
{
    assert( type == METRICTAB );

    clearExternalValues();

    const std::vector<cube::Metric*>& metricVec = cubeExternal->get_metv();
    const std::vector<cube::Cnode*>&  callVec   = cubeExternal->get_cnodev();
    const std::vector<cube::Thread*>& threadVec = cubeExternal->get_thrdv();

    //build the tree of the external metric items
    for ( int i = 0; i < ( int )metricVec.size(); i++ )
    {
        //take the next metric item
        cube::Metric*   metric       = metricVec[ i ];
        cube::Metric*   metricParent = metric->get_parent();
        TreeWidgetItem* itemParent   = NULL;
        //search for the (already created) parent item
        if ( metricParent != NULL )
        {
            for ( int j = 0; j < ( int )externalMetricItems.size(); j++ )
            {
                if ( ( ( cube::Metric* )( ( TreeWidgetItem* )( externalMetricItems[ j ]->cubeObject ) ) ) == metricParent )
                {
                    itemParent = externalMetricItems[ j ];
                    break;
                }
            }
            assert( itemParent != NULL );
        }
        //create a new item for the current metric with the found parent
        TreeWidgetItem* item = new TreeWidgetItem( itemParent );
        item->cubeObject = metric;
        externalMetricItems.push_back( item );
        //we store the metric names because we delete the cube object
        //after computing the external values
        externalMetricNames.push_back( metric->get_uniq_name() );

        //compute external metric values
        double totalValue = 0.0;
        for ( int j = 0; j < ( int )callVec.size(); j++ )
        {
            for ( int k = 0; k < ( int )threadVec.size(); k++ )
            {
                totalValue += cube->get_sev( metric,
                                             callVec[ j ],
                                             threadVec[ k ] );
            }
        }
        item->totalValue = totalValue;
        item->ownValue   = totalValue;
        if ( itemParent != NULL )
        {
            itemParent->ownValue -= totalValue;
        }
    }
    //clean up the cube object references such that the cube
    //object can be deleted
    for ( int i = 0; i < ( int )( externalMetricItems.size() ); i++ )
    {
        externalMetricItems[ i ]->cubeObject = NULL;
    }
}


/***************** display **********************/

// update the view of the current tab element
//
void
TabWidget::displayItems()
{
    if ( count() > 0 )
    {
        if ( type == SYSTEMTAB )
        {   // always update SystemTreeWidget, because following widgets may refer on its values
            tabInterface( systemTreeIndex )->valuesChanged();
            if ( currentIndex() != systemTreeIndex )
            {
                tabInterface( currentIndex() )->valuesChanged();
            }
        }
        else
        {
            tabInterface( currentIndex() )->valuesChanged();
        }
    }
}

/** signal currentChanged() is emitted when the visible tab element has changed;
   that signal gets connected to this slot;
   parameter: index is the new index of the current tab element
 */
void
TabWidget::onCurrentChanged( int index )
{
    SystemTopologyToolBar::getInstance()->setParent( NULL );

    TabInterface* tab = tabInterface( currentIndex() );
    tab->valuesChanged();
    tab->setActive();

    //if current tab element is a tree
    if ( index == 0 || type == CALLTAB )
    {
        computeValues();
    }
}

//defines if zero values in topologies should be represented by the
//color white or by the minimal color;
//this slot is connected to the actions in the menu
//Display/Coloring/Topology coloring

void
TabWidget::whiteOff()
{
    assert( type == SYSTEMTAB );
    whiteForZero = false;
    if ( currentSysTopWidget != NULL )
    {
        currentSysTopWidget->updateColors();
        displayItems();
    }
}
void
TabWidget::whiteOn()
{
    assert( type == SYSTEMTAB );
    whiteForZero = true;
    if ( currentSysTopWidget != NULL )
    {
        currentSysTopWidget->updateColors();
        displayItems();
    }
}

//set line coloring in topology widgets;
//connected to the actions in the menu Display/Coloring/Topology line coloring

void
TabWidget::setLineType( LineType lineType )
{
    assert( type == SYSTEMTAB );
    this->lineType = lineType;
    for ( int i = 0; i < count(); i++ )
    {
        tabInterface( i )->setLineType( lineType );
    }
}
void
TabWidget::blackLines()
{
    setLineType( BLACK_LINES );
}
void
TabWidget::grayLines()
{
    setLineType( GRAY_LINES );
}
void
TabWidget::whiteLines()
{
    setLineType( WHITE_LINES );
}
void
TabWidget::noLines()
{
    setLineType( NO_LINES );
}


//define if lines in topologies should be drawn with antialiasing
//
void
TabWidget::setAntialiasing( bool value )
{
    antialiasing = value;
    displayItems();
}

//toggle modus if lines in topologies should be drawn with antialiasing
void
TabWidget::toggleAntialiasing()
{
    assert( type == SYSTEMTAB );
    antialiasing = !antialiasing;
    displayItems();
}

//change the modus if unused topology planes should be displayed
void
TabWidget::toggleUnused()
{
    assert( type == SYSTEMTAB );
    showUnusedTopologyPlanes = !showUnusedTopologyPlanes;
    for ( int i = 0; i < count(); i++ )
    {
        tabInterface( i )->toggleUnused();
    }
}



/***************** settings ***********************/

QString
TabWidget::typeName()
{
    QString name;
    switch ( type )
    {
        case METRICTAB:
            name = "metrictab";
            break;
        case CALLTAB:
            name = "calltab";
            break;
        case SYSTEMTAB:
            name = "systemtab";
            break;
        default:
            assert( false );
    }
    return name;
}

//save tab settings
//
void
TabWidget::saveSettings( QSettings& settings )
{
    assert( count() > 0 );

    QString groupName = typeName();
    settings.beginGroup( groupName );

    // save settings for each tab
    for ( int i = 0; i < count(); i++ )
    {
        tabInterface( i )->saveSettings( settings );
    }

    //miscellaneous
    settings.setValue( "currentIndex", currentIndex() );
    settings.setValue( "selectionSyntax", selectionSyntax );
    settings.setValue( "whiteForZero", whiteForZero );
    if ( type == SYSTEMTAB )
    {
        settings.setValue( "lineType", lineType );
    }

    settings.endGroup();
}

//load tab settings
//
bool
TabWidget::loadSettings( QSettings& settings )
{
    assert( count() > 0 );

    bool result = true;

    QString groupName = typeName();
    settings.beginGroup( groupName );

    // load settings for each tab
    result = true;
    for ( int i = 0; i < count() && result; i++ )
    {
        result = tabInterface( i )->loadSettings( settings );
    }

    //miscellaneous
    setCurrentIndex( settings.value( "currentIndex", 0 ).toInt() );

    selectionSyntax = ( SelectionSyntax )( settings.value( "selectionSyntax", 0 ).toInt() );
    whiteForZero    = settings.value( "whiteForZero", true ).toBool();
    if ( currentSysTopWidget != NULL )
    {
        currentSysTopWidget->updateColors();
    }

    if ( type == SYSTEMTAB )
    {
        LineType lineType;
        lineType = ( LineType )( settings.value( "lineType", 0 ).toInt() );
        this->setLineType( lineType );
    }

    settings.endGroup();

    displayItems();

    return result;
}



/************ miscellaneous **************/

//return the list of external metric items
int
TabWidget::numExternalMetricItems()
{
    return externalMetricItems.size();
}

QSize
TabWidget::sizeHint()
{
    return tabInterface( currentIndex() )->widget()->sizeHint();
}

/**
   enables box plot tab, if system tab is not first tab
 */
void
TabWidget::enableSystemBoxPlot( bool first )
{
    if ( ( systemBoxPlot ) && ( type == SYSTEMTAB ) )
    {
        setTabEnabled( systemBoxPlotIndex, !first );
    }
}
