/****************************************************************************
**  CUBE        http://www.score-p.org/                                    **
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2014                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/



#include <cassert>

#include "TreeWidgetItem.h"
#include "MyTreeWidget.h"

TreeWidgetItem::TreeWidgetItem( TreeWidgetItem* parent )
{
    name = "";

    framed                    = false;
    statistic_available_frame = false;
    maxItem                   = true;

    id         = 0;
    type       = METRICITEM;
    cubeObject = NULL;

    ownValue   = 0.0;
    totalValue = 0.0;
    rootValue  = 0.0;

    ownValue_adv   = NULL;
    totalValue_adv = NULL;
    rootValue_adv  = NULL;
    singleValue    = true;
    //the minimal and maximal peer values for system tree items
    minValue = 0.0;
    maxValue = 0.0;

    children.clear();

    expanded = false;
    hidden   = false;
    selected = false;
    visible  = ( parent == NULL );

    textExpanded   = "";
    textCollapsed  = "";
    valueExpanded  = 0.0;
    valueCollapsed = 0.0;
    colorExpanded  = Qt::white;
    colorCollapsed = Qt::white;

    parentItem = parent;
    if ( parent != NULL )
    {
        parent->addChild( this );
        depth = parent->getDepth() + 1;
        //set the tree widget
        tree = parent->getTree();
    }
    else
    {
        depth = 0;
        tree  = NULL;
    }

    //found item?
    mark1 = false;
    //parent of found item?
    mark2 = false;

    //  invalidate from the beginning.
    setInvalid();
}

TreeWidgetItem::~TreeWidgetItem()
{
    foreach( TreeWidgetItem * item, hiddenIterations )
    {
        delete item;
    }

    if ( ownValue_adv != NULL )
    {
        ownValue_adv->Free();
    }
    if ( totalValue_adv != NULL )
    {
        totalValue_adv->Free();
    }
    if ( rootValue_adv != NULL )
    {
        rootValue_adv->Free();
    }
}


/***************** set methods ************************/

//set the expansion state
//
void
TreeWidgetItem::setExpanded( bool expanded )
{
    this->expanded = expanded;
}

//set the hidden state
//
void
TreeWidgetItem::setHidden( bool hidden )
{
    this->hidden = hidden;
}

//set the selected state
//
void
TreeWidgetItem::setSelected( bool selected, bool updateTree )
{
    if ( selected != this->selected )
    {
        this->selected = selected;
        //inform the tree widget that this item was inserted
        if ( tree != NULL && updateTree )
        {
            if ( selected )
            {
                tree->select( this );
            }
            else
            {
                tree->deselect( this );
            }
        }
    }
}

//set the text to be displayed in the expansion state "expanded" to
//"text"
//
void
TreeWidgetItem::setText( bool expanded, QString text )
{
    if ( expanded )
    {
        textExpanded = text;
    }
    else
    {
        textCollapsed = text;
    }
}

//set the parent item
//
void
TreeWidgetItem::setParent( TreeWidgetItem* parent )
{
    parentItem = parent;
}

//set the tree widget the item's tree is contained in
//
void
TreeWidgetItem::setTree( MyTreeWidget* tree )
{
    this->tree = tree;
    if ( selected )
    {
        tree->select( this );
    }
    for ( int i = 0; i < ( int )children.size(); i++ )
    {
        children[ i ]->setTree( tree );
    }
}




/***************** get methods **********************/


//get the item's hierarchy depth
//
int
TreeWidgetItem::getDepth()
{
    return depth;
}


// update all depths of subtree of this items
void
TreeWidgetItem::updateDepth()
{
    TreeWidgetItem* _myparent = parent();
    int             _depth    = 0;
    while ( _myparent != NULL )
    {
        _depth++;
        _myparent = _myparent->parent();
    }
    setDepth( _depth );
}

// updatre all depths of subtree of this items
void
TreeWidgetItem::setDepth( int _depth )
{
    depth = _depth;
    for ( int i = 0; i < childCount(); i++ )
    {
        child( i )->setDepth( depth + 1 );
    }
}




//is the item expanded/hidden/selected?
//
bool
TreeWidgetItem::isExpanded()
{
    return expanded;
}
bool
TreeWidgetItem::isHidden()
{
    return hidden;
}
bool
TreeWidgetItem::isSelected()
{
    return selected;
}

//get the text to be displayed in the current state
//
QString
TreeWidgetItem::getText()
{
    return expanded ? textExpanded : textCollapsed;
}

//get the text to be displayed in the expanded state
//if "expanded" is true, and in the collapsed state otherwise
//
QString
TreeWidgetItem::getText( bool expanded )
{
    return expanded ? textExpanded : textCollapsed;
}

//get the item's value in the current state
//
double
TreeWidgetItem::getValue()
{
    return expanded ? valueExpanded : valueCollapsed;
}

//get the color for the value of the item's current state
//
QColor
TreeWidgetItem::getColor()
{
    return expanded ? colorExpanded : colorCollapsed;
}

//return number of children
//
int
TreeWidgetItem::childCount()
{
    return children.size();
}

//return the ith child
//
TreeWidgetItem*
TreeWidgetItem::child( int index )
{
    if ( ( int )children.size() <= index )
    {
        return NULL;
    }
    else
    {
        return children[ index ];
    }
}

//return the parent item
//
TreeWidgetItem*
TreeWidgetItem::parent()
{
    return parentItem;
}

//get the tree widget the item's tree in contained in
//
MyTreeWidget*
TreeWidgetItem::getTree()
{
    return tree;
}



/********** add/remove childen *********/

//add a new child to this item
//
void
TreeWidgetItem::addChild( TreeWidgetItem* item )
{
    assert( !item->isSelected() );
    assert( item->parent() == NULL || item->parent() == this );
    bool found = false;
    //check if the parameter item is already in the children list
    for ( int i = 0; i < ( int )children.size(); i++ )
    {
        if ( children[ i ] == item )
        {
            found = true;
            break;
        }
    }
    //if not, then add it as child
    if ( !found )
    {
        children.push_back( item );
    }
    item->setParent( this );
}

#include <QDebug>
/**
 * @brief TreeWidgetItem::hideIterations
 * hides the children of the current item, which are iterations, but shows all
 * grandchildren, which will be merged
 */
void
TreeWidgetItem::hideIterations()
{
    hiddenIterations = getChildren();
    this->getChildren().clear();

    this->name.append( " (" + QString::number( hiddenIterations.size() ) + " iterations)" );

    mergeTrees( this, hiddenIterations );
}

/**
 * @brief TreeWidgetItem::mergeTrees recursively merges iteration items with the same name into
 * new items and appends them to newParent.
 * Only items of the same tree depth which have the same name are merged.
 */
void
TreeWidgetItem::mergeTrees( TreeWidgetItem* newParent, vector<TreeWidgetItem*>& iterations )
{
    QList<TreeWidgetItem* > content;
    QStringList             names;

    foreach( TreeWidgetItem * iteration, iterations )
    {
        foreach( TreeWidgetItem * child, iteration->getChildren() )
        {
            if ( !names.contains( child->name ) )
            {
                content.append( child );
                names.append( child->name );
            }
        }
    }
    if ( content.size() == 0 )
    {
        return;
    }

    foreach( TreeWidgetItem * item, content )
    {
        QString         childName = item->name;
        TreeWidgetItem* newItem   = new TreeWidgetItem( newParent );

        newItem->depth    = newParent->depth + 1;
        newItem->invalid  = true;
        newItem->name     = item->name;
        newItem->hidden   = item->hidden;
        newItem->expanded = item->expanded;
        newItem->id       = item->id;
        newItem->type     = item->type;
        newItem->visible  = item->visible;
        newItem->setText( true,  "- " + item->name );
        newItem->setText( false, "- " + item->name );
        newParent->addChild( newItem );

        vector<TreeWidgetItem*> listOfChildren;
        foreach( TreeWidgetItem * iteration, iterations )
        {
            foreach( TreeWidgetItem * child, iteration->getChildren() )
            {
                if ( child->name == childName )
                {
                    listOfChildren.push_back( child );
                }
            }
        }
        if ( listOfChildren.size() > 0 )
        {
            mergeTrees( newItem, listOfChildren );
        }
    }
}

/**
 * show all iterations and its subtrees
 */
void
TreeWidgetItem::unhideIterations()
{
    this->getChildren().clear();

    QString append = " (" + QString::number( hiddenIterations.size() ) + " iterations)";
    this->name.remove( append );

    foreach( TreeWidgetItem * item, hiddenIterations )
    {
        item->setParent( 0 );
        this->addChild( item );
        item->setExpanded( false );

        foreach( TreeWidgetItem * child, item->getChildren() )
        {
            child->setParent( item );
        }
    }
    hiddenIterations.clear();
}

//remove a child from the item's children list
//
void
TreeWidgetItem::removeChild( TreeWidgetItem* item )
{
    assert( item->parent() == this );
    for ( int i = 0; i < ( int )children.size(); i++ )
    {
        if ( children[ i ] == item )
        {
            for (; i < ( int )children.size() - 1; i++ )
            {
                children[ i ] = children[ i + 1 ];
            }
            children.pop_back();
            break;
        }
        assert( i < ( int )children.size() - 1 );
    }
    item->setParent( NULL );
}

void
TreeWidgetItem::setValues( cube::Value* total, cube::Value* own )
{
    if ( totalValue_adv != NULL )
    {
        totalValue_adv->Free();
    }
    if ( ownValue_adv != NULL )
    {
        ownValue_adv->Free();
    }

    totalValue_adv = total;
    ownValue_adv   = own;
    totalValue     = ( totalValue_adv != NULL ) ? totalValue_adv->getDouble() : 0.;
    ownValue       = ( ownValue_adv != NULL ) ? ownValue_adv->getDouble() : 0.;
    valueCollapsed = totalValue;
    valueExpanded  = ownValue;
    singleValue    = ( ownValue_adv != NULL ) ? ownValue_adv->singleValue() : true;

    // for "Own root percent" value mode
    if ( parentItem != NULL )
    {
        rootValue = parentItem->rootValue;
    }
    else
    {
        rootValue = totalValue;
    }
}
