/****************************************************************************
**  CUBE        http://www.score-p.org/                                    **
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2014                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/



#ifndef _TREEWIDGETITEM_H
#define _TREEWIDGETITEM_H

#include <QTreeWidgetItem>
#include <QBrush>

#include "CubeVertex.h"
#include "CubeValues.h"


#include "Constants.h"



class MyTreeWidget;

//TreeWidgetItem is the type of tree nodes


class TreeWidgetItem
{
public:

    TreeWidgetItem( TreeWidgetItem* parent );

    ~TreeWidgetItem();

    /************* fields *************/

    //name of the item
    QString name;

    //values attached to the item
    double ownValue;
    double totalValue;
    double rootValue;

    bool         singleValue;
    cube::Value* ownValue_adv;
    cube::Value* totalValue_adv;
    cube::Value* rootValue_adv;


    //the minimal and maximal peer values for system tree items
    double minValue;
    double maxValue;

    //item values/colors for the current state; note that valueExpanded
    //can differ from ownValue if the item has hidden children, in which
    //case the total values of the hidden children are added up to the
    //own value
    double valueExpanded, valueCollapsed;
    QColor colorExpanded, colorCollapsed;

    //this flag is used in the aggregate mode "MAX"
    //to mark if the node belongs to the maximal cnode tree
    bool maxItem;

    //the index of the node in the item vector
    //of the treewidget object it is contained in
    unsigned id;

    //the type of the node;
    //needed to distinguish how to handle nodes
    TreeWidgetItemType type;

    //the corresponding cube object
    cube::Vertex* cubeObject;

    //should the item be framed?
    bool framed;

    //should the item be framed if statistic is avialable? (only for cnodes)
    bool statistic_available_frame;


    // indicates, that parent is expanded
    bool visible;
    void
    setVisible()
    {
        visible = true;
    }
    void
    setInvisible()
    {
        visible = false;
    }
    bool
    isVisible()
    {
        return visible;
    }

    //node found by "find items"?
    bool mark1;
    //is there a node in this subtree found by "find items"?
    bool mark2;

    /************* set methods **************/

    //set the item's displayed text (name, value, etc.) for the "expanded" state
    void
    setText( bool    expanded,
             QString text );

    //set and get expansion state
    void
    setExpanded( bool expanded );

    //set and get hidden state
    void
    setHidden( bool hidden );

    //set and get selected state
    void
    setSelected( bool selected,
                 bool updateTree = true );

    //set the tree widget (if any) the item is contained in
    void
    setTree( MyTreeWidget* tree );

    //set parent item
    void
    setParent( TreeWidgetItem* item );

    /********************** get methods ********************/

    //get the item's hierarchy depth
    int
    getDepth();

    //updatre all depths of subtree of this item
    void
    updateDepth();

    // sets depths of itself and children
    void
    setDepth( int depth );

    //get the text to be displayed in the current expansion mode
    QString
    getText();

    //get the text to be displayed in the expansion mode "expanded"
    QString
    getText( bool expanded );

    //get color for the current value of the item
    QColor
    getColor();

    //get current value of the item
    double
    getValue();

    //get the tree widget (if any) the item is contained in
    MyTreeWidget*
    getTree();

    //get parent item
    TreeWidgetItem*
    parent();

    //number of children
    int
    childCount();

    //return the ith child
    TreeWidgetItem*
    child( int index );

    // returns a vector holding all children
    vector<TreeWidgetItem*>&
    getChildren()
    {
        return children;
    };


    //is the item expanded/hidden/selected?
    bool
    isExpanded();
    bool
    isHidden();
    bool
    isSelected();

    /********** add/remove children *********/

    //remove a child
    void
    removeChild( TreeWidgetItem* item );

    //add a new child
    void
    addChild( TreeWidgetItem* item );


    // sets the Invalide flag,
    void
    setInvalid()
    {
        invalid = true;
    };                                                      ///< makes item invalid and to be recalculated
    void
    setValid()
    {
        invalid = false;
    };                                                      ///< makes item valid

    void
    setStatisticFlag()
    {
        statistic_available_frame = true;
    }
    void
    clearStatisticFlag()
    {
        statistic_available_frame = false;
    }

    bool
    isInvalid()
    {
        return invalid;
    };

    void
    ValueIsNonValue( bool isNaN )
    {
        valueIsNonValue = isNaN;
    };                                                                      ///< sets "NaN" flag

    bool
    singleValued()
    {
        return singleValue;
    };                                            ///< Signalizesm, that the saved value is a single integer.

    void
    hideIterations();
    void
    unhideIterations();
    void
    setValues( cube::Value* total,
               cube::Value* own );

    const vector<TreeWidgetItem*>&
    getHiddenIterations()
    {
        return hiddenIterations;
    }

private:
    void
    mergeTrees( TreeWidgetItem*          newParent,
                vector<TreeWidgetItem*>& iterations );

    vector<TreeWidgetItem*> hiddenIterations;
    bool                    expanded;
    bool                    hidden;
    bool                    selected;

    TreeWidgetItem*              parentItem;
    std::vector<TreeWidgetItem*> children;

    int           depth;
    MyTreeWidget* tree;

    QString textExpanded, textCollapsed;


    // -------------- adv part -----------
    bool invalid;               ///< signalizes, that the value of its item has to be recalculated

    bool valueIsNonValue;       ///< signalizes, that the value, carried by this item was involved in the arithmetic with "NonValue"
};

#endif
