/****************************************************************************
**  CUBE        http://www.score-p.org/                                    **
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2014                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <cassert>
#include <string>

#include <QHBoxLayout>
#include <QPushButton>
#include <QMessageBox>

#include "TreeWidget.h"
#include "MessageWidget.h"

#include "Statistics.h"
#include "BoxPlot.h"

#include "Cube.h"
#include "CubeMetric.h"
#include "CubeCnode.h"

using namespace std;

void
TreeWidget::createStatisticsContextMenuItems()
{
    //for metric trees create an action for expert statistics
    //and for connections to a trace browser
    if ( statistics != 0 && statistics->existsStatFile() )
    {
        if ( type == METRICTREE )
        {
            contextMenu.addSeparator();

            statisticsAct = new Action( tr( "Statistics" ), this );
            statisticsAct->setStatusTip( tr( "Shows metric statistics" ) );
            statisticsAct->setWhatsThis( tr( "Shows metric statistics. Only available if a statistics file for the current cube file exists and if statistical information for the selected metric is provided." ) );
            connect( statisticsAct, SIGNAL( triggered() ), this, SLOT( onShowStatistics() ) );
            contextMenu.addAction( statisticsAct );
        }
        if ( type == METRICTREE || type == CALLTREE )
        {
            traceActText = new Action( tr( "Show max severity information" ), this );
            traceActText->setStatusTip( tr( "Shows the most severe instance of pattern as text" ) );
            traceActText->setWhatsThis( tr( "Shows the most severe instance of pattern as text. Only available if a statistics file for the current cube file exists." ) );
            connect( traceActText, SIGNAL( triggered() ), this, SLOT( onShowMaxSeverityText() ) );
            contextMenu.addAction( traceActText );
            if ( !statistics->existsStatFile() )
            {
                traceActText->setVisible( false );
            }

#if defined( WITH_VAMPIR ) || defined( WITH_PARAVER )
            traceAct = new Action( tr( "Max severity in trace browser" ), this );
            traceAct->setStatusTip( tr( "Shows the most severe instance of pattern in trace browser" ) );
            traceAct->setWhatsThis( tr( "Shows the most severe instance of pattern in trace browser. Only available if a statistics file for the current cube file exists and if a connection to a trace browser was established." ) );
            connect( traceAct, SIGNAL( triggered() ), this, SLOT( onShowMaxSeverity() ) );
            contextMenu.addAction( traceAct );
            if ( !statistics->existsStatFile() )
            {
                traceAct->setVisible( false );
            }
#endif
        }
    }
}

void
TreeWidget::onShowStatistics()
{
    if ( statistics == 0 )
    {
        return;                    // ignore , and don't not break
    }
    if ( clickedItem == NULL )
    {
        return;                        // ignore, don't crash
    }
    if ( type != METRICTREE )
    {
        return;                       // ignore, don't exit
    }
    QDialog* dialog = new QDialog( this );
    dialog->setAttribute( Qt::WA_DeleteOnClose );
    dialog->setWindowTitle( "Statistics info" );
    BoxPlot* theBoxPlot = new BoxPlot( dialog, statistics->getPrecisionWidget() );

    vector<TreeWidgetItem*> const& vec   = selectedItems();
    bool                           added = false;
    for ( vector<TreeWidgetItem*>::const_iterator it = vec.begin(); it != vec.end(); ++it )
    {
        if ( statistics->existsStatistics( static_cast<cube::Metric*>( ( *it )->cubeObject ) ) )
        {
            theBoxPlot->Add( statistics->getItem( static_cast<cube::Metric*>( ( *it )->cubeObject ) ) );
            if ( *it == clickedItem )
            {
                added = true;
            }
        }
    }
    if ( !added )
    {
        theBoxPlot->Add( statistics->getItem( static_cast<cube::Metric*>( clickedItem->cubeObject ) ) );
    }

    QVBoxLayout* layout = new QVBoxLayout();
    layout->addWidget( theBoxPlot );

    QPushButton* okButton = new QPushButton( "Close", dialog );
    connect( okButton, SIGNAL( pressed() ), dialog, SLOT( accept() ) );
    layout->addWidget( okButton );

    dialog->setLayout( layout );

    dialog->setModal( false );
    dialog->show();
}

void
TreeWidget::onShowMaxSeverityText()
{
    assert( statistics != 0 );
    assert( clickedItem != 0 );
    assert( type == METRICTREE || type == CALLTREE );

    TreeWidgetItem*     metricItem = tabWidget->getMetricTree()->selectedItems().at( 0 );
    cube::Metric const* metric     = static_cast<cube::Metric*>( metricItem->cubeObject );
    TreeWidgetItem*     callItem   = 0;

    if ( type == METRICTREE )
    {
        uint32_t cnodeID = statistics->findMaxSeverityId( metric );

        // find call item with given cnodeID
        TreeWidget*                  callTree = tabWidget->getCallTree();
        std::vector<TreeWidgetItem*> items    = callTree->items;
        for ( std::vector<TreeWidgetItem*>::iterator it = items.begin(); it != items.end(); ++it )
        {
            TreeWidgetItem* item = *it;
            if ( item->cubeObject->get_id() == cnodeID )
            {
                callItem = item;
                break;
            }
        }
    }
    else if ( type == CALLTREE )
    {
        callItem = clickedItem;
    }

    QString path;

    TreeWidgetItem* item = callItem;
    while ( item != 0 )
    {
        QString sep  = item->getDepth() > 0 ? "+ " : "";
        QString line = QString( 2 * item->getDepth(), ' ' ) + sep + item->name;
        path = ( item == callItem ) ? line : line + "\n" + path;
        item = item->parent();
    }
    QString label = QString( "metric: \n  " ) + metric->get_descr().c_str();
    label += QString( " [" ) + metric->get_uom().c_str() + "]";
    if ( type == METRICTREE )
    {
        label += "\ncallpath of max severity: \n";
    }
    else if ( type == CALLTREE )
    {
        label += "\nselected callpath: \n";
    }
    label += path;

    cube::Cnode const* cnode = static_cast<cube::Cnode*>( callItem->cubeObject );
    statistics->showMaxSeverityText( this, label, metric, cnode );
}

#if defined( WITH_VAMPIR ) || defined( WITH_PARAVER )
void
TreeWidget::onShowMaxSeverity()
{
    assert( statistics != 0 );
    assert( statistics->isConnected() );
    assert( clickedItem != 0 );
    assert( type == METRICTREE || type == CALLTREE );
    string str;
    if ( type == METRICTREE )
    {
        str = statistics->showMaxSeverity( static_cast<cube::Metric*>( clickedItem->cubeObject ) );
    }
    else if ( type == CALLTREE )
    {
        cube::Metric* metric = static_cast<cube::Metric*>( tabWidget->getMetricTree()->selectedItems().at( 0 )->cubeObject );
        str = statistics->showMaxSeverity( metric, static_cast<cube::Cnode*>( clickedItem->cubeObject ) );
    }
    if ( str != "" )
    {
        MessageWidget* wwidget = new MessageWidget( Error, str.c_str() );
        wwidget->show();
//          QMessageBox::critical(this, "Statistics", str.c_str());
    }
}
#endif

void
TreeWidget::statisticsMousePressEvent()
{
    //the statistics action serves with some expert statistics if available
    if ( clickedItem == NULL )
    {
        return;
    }
    if ( statistics != 0 && statistics->existsStatFile() )
    {
        if ( type == METRICTREE )
        {
            cube::Metric* metric = static_cast<cube::Metric*>( clickedItem->cubeObject );

            bool enableStatistics = false;
            if ( clickedItem != 0 && statistics != 0
                 && statistics->existsStatistics( metric ) )
            {
                enableStatistics = true;
            }
            statisticsAct->setEnabled( enableStatistics );
        }

        //the trace action offers the ability to connect to let the trace
        //browser show the most severe instance of a pattern
        if ( type == METRICTREE )
        {
            traceActText->setEnabled( false );
#if defined( WITH_VAMPIR ) || defined( WITH_PARAVER )
            traceAct->setEnabled( false );
#endif
            if ( clickedItem != 0 && statistics != 0
                 && statistics->existsMaxSeverity( static_cast<cube::Metric*>( clickedItem->cubeObject ) ) )
            {
                traceActText->setEnabled( true );
#if defined( WITH_VAMPIR ) || defined( WITH_PARAVER )
                if ( statistics->isConnected() )
                {
                    traceAct->setEnabled( true );
                }
#endif
            }
        }
        if ( type == CALLTREE )
        {
            traceActText->setEnabled( false );
#if defined( WITH_VAMPIR ) || defined( WITH_PARAVER )
            traceAct->setEnabled( false );
#endif

            cube::Metric* metric = static_cast<cube::Metric*>( tabWidget->getMetricTree()->selectedItems().at( 0 )->cubeObject );
            if ( clickedItem != 0 && statistics != 0 && tabWidget->getMetricTree()->selectedItems().size() == 1
                 && statistics->existsMaxSeverity( metric, static_cast<cube::Cnode*>( clickedItem->cubeObject ) ) )
            {
                traceActText->setEnabled( true );
#if defined( WITH_VAMPIR ) || defined( WITH_PARAVER )
                if ( statistics->isConnected() )
                {
                    traceAct->setEnabled( true );
                }
#endif
            }
        }
    }
}





void
TreeWidget::updateStatisticsSelection()
{
    if ( statistics != 0 && statistics->existsStatFile() )
    {
        if ( type == METRICTREE || type == CALLTREE )
        {
            std::vector<TreeWidgetItem*> metrics = tabWidget->getMetricTree()->selectedItems();

            std::vector<TreeWidgetItem*> cnodes_with_stat;
            std::vector<TreeWidgetItem*> cnodes = tabWidget->getCallTree()->Items();

            for ( std::vector<TreeWidgetItem*>::iterator c_iter = cnodes.begin();
                  c_iter != cnodes.end(); c_iter++   )
            {
                ( *c_iter )->clearStatisticFlag();
            }

            for ( std::vector<TreeWidgetItem*>::iterator m_iter = metrics.begin();
                  m_iter != metrics.end(); m_iter++   )
            {
                cube::Metric* metric = static_cast<cube::Metric*>( ( *m_iter )->cubeObject );

                for ( std::vector<TreeWidgetItem*>::iterator c_iter = cnodes.begin();
                      c_iter != cnodes.end(); c_iter++   )
                {
                    cube::Cnode* cnode = ( static_cast<cube::Cnode*>( ( *c_iter )->cubeObject ) );

                    if ( statistics->existsMaxSeverity( metric, cnode ) )
                    {
                        cnodes_with_stat.push_back( *c_iter );
                    }
                }
            }
            for ( std::vector<TreeWidgetItem*>::iterator c_iter = cnodes_with_stat.begin();
                  c_iter != cnodes_with_stat.end(); c_iter++   )
            {
                TreeWidgetItem* _item = *c_iter;
                while ( !_item->isVisible() )
                {
                    _item = _item->parent();
                }

                _item->setStatisticFlag();
            }
            tabWidget->getCallTree()->update();
        }
    }
}
