/****************************************************************************
**  CUBE        http://www.score-p.org/                                    **
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2014                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#ifndef VAMPIR_CONNECTION_DIALOG_CPP
#define VAMPIR_CONNECTION_DIALOG_CPP

#include <QDialog>
#include <QLineEdit>
#include <QPushButton>
#include <QSpinBox>
#include <QVBoxLayout>
#include <QGridLayout>
#include <QLabel>
#include <QDialogButtonBox>
#include <cassert>

#include "MainWidget.h"
#include "MessageWidget.h"
#include "Statistics.h"
#include "VampirConnectionDialog.h"

class VampirConnectionThread : public QThread
{
public:
    VampirConnectionThread( string const&           host,
                            int                     port,
                            string const&           file,
                            VampirConnectionDialog* dia,
                            Statistics*             stat );
    virtual void
    run();

private:
    string                  hostName;
    string                  fileName;
    int                     portNumber;
    Statistics*             statistics;
    VampirConnectionDialog* dialog;
};

VampirConnectionThread::VampirConnectionThread( string const&           host,
                                                int                     port,
                                                string const&           file,
                                                VampirConnectionDialog* dia,
                                                Statistics*             stat )
    : hostName( host ), fileName( file ), portNumber( port ), statistics( stat ),
      dialog( dia )
{
}

void
VampirConnectionThread::run()
{
    string result = statistics->connectToVampir( hostName, portNumber, fileName );
    if ( result != "" )
    {
        dialog->setError( "Vampir connection", result );
    }
}

VampirConnectionDialog::VampirConnectionDialog( QWidget*    par,
                                                QLineEdit*& fileEdit,
                                                Statistics* stat,
                                                MainWidget* mainWid )
    : QDialog( par ), statistics( stat ), connectionThread( 0 )
{
    int const spacing = 20;

    setWindowTitle( "Connect to vampir" );

    QFontMetrics fm( font() );
    int          h = 10 * fm.ascent();
    int          w = 4 * fm.width( "Connect to vampir:" );
    setMinimumSize( w, h );

    QVBoxLayout* layout = new QVBoxLayout();
    layout->setSpacing( spacing );

    QGridLayout* layout1 = new QGridLayout();
    layout1->setSpacing( spacing );

    //define host label and editor
    QLabel* hostLabel = new QLabel();
    hostLabel->setText( "Host:" );
    hostLabel->setDisabled( true );
    layout1->addWidget( hostLabel, 1, 0 );

    hostLine = new QLineEdit( "localhost" );
    hostLine->setDisabled( true );
    layout1->addWidget( hostLine, 1, 1 );

    //define port label and editor
    QLabel* portLabel = new QLabel();
    portLabel->setText( "Port:" );
    portLabel->setDisabled( true );
    layout1->addWidget( portLabel, 2, 0 );

    portLine = new QSpinBox();
    portLine->setRange( 0, 1000000 );
    portLine->setSingleStep( 1 );
    portLine->setValue( 30000 );
    portLine->setDisabled( true );
    layout1->addWidget( portLine, 2, 1 );

    //set checkbox to change between a local or remote trace file
    fromServerCheckbox = new QCheckBox( "Open local file" );
    fromServerCheckbox->setChecked( true );
    layout1->addWidget( fromServerCheckbox, 0, 0, 1, 2 );
    connect( fromServerCheckbox, SIGNAL( toggled( bool ) ), portLine, SLOT( setDisabled( bool ) ) );
    connect( fromServerCheckbox, SIGNAL( toggled( bool ) ), hostLine, SLOT( setDisabled( bool ) ) );
    connect( fromServerCheckbox, SIGNAL( toggled( bool ) ), portLabel, SLOT( setDisabled( bool ) ) );
    connect( fromServerCheckbox, SIGNAL( toggled( bool ) ), hostLabel, SLOT( setDisabled( bool ) ) );

    layout->addLayout( layout1 );

    //define file label and editor

    QHBoxLayout* layout2 = new QHBoxLayout();
    layout2->setSpacing( spacing );

    QLabel* fileLabel = new QLabel();
    fileLabel->setText( "File:" );
    layout2->addWidget( fileLabel );

    fileEdit = fileLine =
                   new QLineEdit( statistics->getDefaultVampirFileName().c_str() );
    layout2->addWidget( fileLine );
    fileButton = new QPushButton( "Browse" );
    layout2->addWidget( fileButton );
    connect( fileButton, SIGNAL( clicked() ), mainWid, SLOT( getTraceFileName() ) );

    layout->addLayout( layout2 );

    QDialogButtonBox* buttonBox = new QDialogButtonBox();
    okButton     = buttonBox->addButton( QDialogButtonBox::Ok );
    cancelButton = buttonBox->addButton( QDialogButtonBox::Cancel );
    connect( okButton, SIGNAL( clicked() ), this, SLOT( establishVampirConnection() ) );
    connect( cancelButton, SIGNAL( clicked() ), this, SLOT( reject() ) );
    layout->addWidget( buttonBox );

    setLayout( layout );
}

void
VampirConnectionDialog::setError( string const& title, string const& message )
{
    errorTitle   = title;
    errorMessage = message;
}

void
VampirConnectionDialog::printError()
{
    if ( errorMessage != "" )
    {
        MessageWidget* wwidget = new MessageWidget( Error, QString::fromStdString( errorTitle ).append( "\n" ).append( QString::fromStdString( errorMessage ) ) );
        wwidget->show();

//         QMessageBox::critical(this, QString::fromStdString(errorTitle),
//                               QString::fromStdString(errorMessage));
    }
    setAttribute( Qt::WA_DeleteOnClose );
    accept();
}

void
VampirConnectionDialog::establishVampirConnection()
{
    okButton->setEnabled( false );
    cancelButton->setEnabled( false );
    fileButton->setEnabled( false );

    assert( statistics != 0 );
    if ( fromServerCheckbox->isChecked() )
    {
        connectionThread =
            new VampirConnectionThread( "", 0,
                                        fileLine->displayText().toStdString(), this,
                                        statistics );
    }
    else
    {
        connectionThread =
            new VampirConnectionThread( hostLine->displayText().toStdString(),
                                        portLine->value(), fileLine->displayText().toStdString(), this,
                                        statistics );
    }
    connect( connectionThread, SIGNAL( finished() ), this, SLOT( printError() ) );
    connectionThread->start();
}

VampirConnectionDialog::~VampirConnectionDialog()
{
    delete connectionThread;
}
#endif
