/****************************************************************************
**  CUBE        http://www.score-p.org/                                    **
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2014                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2014                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/

/**
 * \file CubeMetric.h
 * \brief Defines a metric class as a element of the metric dimension. It doesn't make any assumptotion about relations between part-child entries. Inclusive and Exclusive Value returns jast saved element.
 *
 */
#ifndef CUBE_METRIC_H
#define CUBE_METRIC_H


#include <iosfwd>
#include <map>

#include "CubeVertex.h"
#include "CubeTypes.h"
#include "CubeCnode.h"
#include "CubeThread.h"
#include "CubeProcess.h"
#include "CubeNode.h"
#include "CubeMachine.h"

//#include "Matrix.h"
#include "CubeRowWiseMatrix.h"
#include "CubeIDdeliverer.h"
#include "CubeFileFinder.h"
#include "CubeStrategies.h"
#include "CubeObjectsEnumerator.h"

#include "CubeCaches.h"

#include "CubePLMemoryManager.h"
#include "CubeGeneralEvaluation.h"
/*
   *----------------------------------------------------------------------------
 *
 * class Metric
 *
   =======
 */


namespace cube
{
enum TypeOfMetric { CUBE_METRIC_EXCLUSIVE,
                    CUBE_METRIC_INCLUSIVE,
                    CUBE_METRIC_SIMPLE,
                    CUBE_METRIC_POSTDERIVED,
                    CUBE_METRIC_PREDERIVED_INCLUSIVE,
                    CUBE_METRIC_PREDERIVED_EXCLUSIVE,
                    CUBE_METRIC_GHOST,
                    CUBE_METRIC_NO_GHOST };
// enum MetricCalculationFlavour { CUBE_METRIC_CALCULATE_INCLUSIVE, CUBE_METRIC_CALCULATE_EXCLUSIVE };





/*  class Cnode;
   class Thread;*/
class GeneralEvaluation; // forward declaration;

/**
 * Metric has a treelike structure and is a kind of Vertex.
 *
 * Metric has a display name, unique name, data type, unit of measure (sec, ms and so on), url to the documentation, description of the metric, ID.
 */
class Metric : public Vertex
{
public:

    Metric( const std::string& disp_name,
            const std::string& uniq_name,
            const std::string& dtype,
            const std::string& uom,
            const std::string& val,
            const std::string& url,
            const std::string& descr,
            FileFinder*        ffinder,
            Metric*            parent,
            uint32_t           id = 0,
            const std::string& _expression = "",
            const std::string& _expression_init = "",
            TypeOfMetric       is_ghost = CUBE_METRIC_NO_GHOST
            )
        : Vertex( parent, id ),
          disp_name( disp_name ),
          uniq_name( uniq_name ),
          dtype( dtype ),
          uom( uom ),
          val( val ),
          url( url ),
          descr( descr ),
          filefinder( ffinder ),
          expression( _expression ),
          expression_init( _expression_init ),
          is_ghost( is_ghost )
    {
        own_hash              = construct_own_hash();
        own_data_type         = get_data_type();
        adv_sev_mat           = NULL;
        cache                 = NULL;
        initialized           = false;
        evaluation            = NULL;
        init_evaluation       = NULL;
        cubepl_memory_manager = NULL;
        active                = true;
        my_strategy           = cube::CUBE_ALL_IN_MEMORY_STRATEGY;
        metric_value          = selectValueOnDataType( own_data_type );
    }
    virtual
    ~Metric();

/**
 * Returns the i-th "submetric"
 */
    Metric*
    get_child( unsigned int i ) const
    {
        return static_cast<Metric*>( Vertex::get_child( i ) );
    }
/**
 * This metric is a submetric of ...?
 */
    Metric*
    get_parent()     const
    {
        return static_cast<Metric*>( Vertex::get_parent() );
    }

/**
 * Return metric value characteristics (such as "VOID") if optionally set,
   or value characteristics inherited from ancestors.
 */
    std::string
    get_val() const
    {
        Metric* _parent = get_parent();
        if ( val != "" || ( _parent == NULL ) )
        {
            return val;
        }
        else
        {
            return _parent->get_val();
        }
    }
    static TypeOfMetric
    get_type_of_metric( std::string s_kind );

    virtual TypeOfMetric
    get_type_of_metric() const
    {
        return CUBE_METRIC_SIMPLE;
    };
    virtual std::string
    get_metric_kind() const
    {
        return "";
    }
    std::string
    get_disp_name() const
    {
        return disp_name;
    }                                                                   ///< Displayed name
    std::string
    get_uniq_name() const
    {
        return uniq_name;
    }                                                                   ///< Unique name
    std::string
    get_dtype()     const
    {
        return dtype;
    }                                                                   ///< Type of the metric
    std::string
    get_uom()       const
    {
        return uom;
    }                                                                   ///< Unit of measurement
    std::string
    get_descr()     const
    {
        return descr;
    }                                                                   ///< Description of metric
    std::string
    get_url()       const
    {
        return url;
    }                                                                   ///< URL for metric details
    std::string
    get_expression()       const
    {
        return expression;
    }                                                                   ///< URL for metric details

    void
    set_expression( std::string _expression )                            ///< Sets the new expression into the metric. No compilation or check is perfomed.
    {
        expression = _expression;
    }
    std::string
    get_init_expression()       const
    {
        return expression_init;
    }                                                                   ///< URL for metric details

    void
    set_init_expression( std::string _expression )                            ///< Sets the new expression into the metric. No compilation or check is perfomed.
    {
        expression_init = _expression;
    }

    void
    setMemoryManager( CubePLMemoryManager* _manager );                                                     ///< Cube parses expression and creates an hierarchy of evaluators, which is saved here


    GeneralEvaluation*
    getEvaluation()                                                     ///< Returns an object to evaluate the value of the metric
    {
        return evaluation;
    }


    void
    setEvaluation( GeneralEvaluation* _ev );                                                     ///< Cube parses expression and creates an hierarchy of evaluators, which is saved here




    GeneralEvaluation*
    getInitEvaluation()                                                     ///< Returns an object to evaluate the value of the metric
    {
        return init_evaluation;
    }
    void
    setInitEvaluation( GeneralEvaluation* _ev );

    void
    set_val( const std::string& value )
    {
        val = value;
    }                                                                   ///< set value type




    double
    get_sev( Cnode*  cnode,
             Thread* thrd );                                            ///< returns saved value, without taking into account, is it exclusive value or inclusive. (Backward compatibility... )
    Value*
    get_sev_adv( Cnode*  cnode,
                 Thread* thrd );                                             ///< returns saved value, without taking into account, is it exclusive value or inclusive.

    void
    set_sev( Cnode*  cnode,
             Thread* thrd,
             double  value );                                           ///< saves value in Metric.                                                                (Backward compatibility ...)

    double
    sum();                                                              ///< Returns sum of all data in the metric. [ Taken fror cube3... not adapted to Cube4 yet]
    void
    clear();                                                            ///< Clears all the data in the metric.     [ Taken fror cube3... not adapted to Cube4 yet]


    virtual bool
    weakEqual( Metric* );

    virtual bool
    weakEqual( Vertex* );


    virtual std::string
    construct_own_hash()
    {
        return uniq_name;
    };                                                                  ///< Uniq name per definitio is a hash.

    /// writes the metric definition into fstream (writes child metrics recursively)
    void
    writeXML( std::ostream& out,
              bool          cube3_export = true ) const;

    /// writes the data into fstream. Produces cube3 layout.
    void
    writeXML_data( std::ostream&         out,
                   std::vector<Cnode*>&  cnodev,
                   std::vector<Thread*>& thrdv );



    // ---------- adv Data start ------------------------
    void
    writeData();                                                              ///< Performs actual final of the data to disk (Of cached data container - then last line gets saved)

    virtual
    void create_data_container( uint32_t, uint32_t );                         ///< Creates data container, in the case, the cube will lbe created.
    virtual
    void create_existing_data_container( uint32_t, uint32_t );                ///< Creates data container, in the case, the cube is existing, (Readonly modus)
    virtual row_of_objects_t*
    create_calltree_id_maps( IDdeliverer*,
                             Cnode*,
                             row_of_objects_t*  );                                              ///< Creates apropriate enumeration of the calltree. Exclusive metric need deep search, Inclusive -> wide search enumeration.


    void
    set_dimensions( std::vector<Cnode*>&,
                    std::vector<Cnode*>&,
                    std::vector<Machine*>&,
                    std::vector<Thread*>& );                                                                     ///< saves inside of the object Metric information about another dimensions, their roots and flat list of elements.


    bool
    isInitialized()
    {
        return initialized;
    };                                                                            ///< Metric internal flag, which invokes initialization by first data acces .

    bool
    isInactive()
    {
        return !active;
    };                                                                          ///< Metrics internal flag, whiche signalizes,, that metric is not being active and should be not stored.


    void
    setActive( bool _flag )
    {
        active = _flag;
    };                                                                          ///< Set the metric active or inactive.

    bool
    isGhost()
    {
        return is_ghost == CUBE_METRIC_GHOST;
    }

    inline
    virtual bool
    asInclusiveMetric()
    {
        return metric_value->asInclusiveMetric();
    }

    inline
    virtual bool
    asExclusiveMetric()
    {
        return metric_value->asExclusiveMetric();
    }

    /// Returns true if the value is INTEGER
    bool
    isIntegerType();


    DataType
    get_data_type();                                                                ///< Returns the data type of this metric.
    virtual IndexFormat
    get_index_format();                                                             ///< Returns format of the index, used in this metric (used ... where?)

    void
    set_adv_sev( Cnode*  cnode,
                 Thread* thrd,
                 double  value );                                                   ///< Saves the double value into the underlying AdvancedMatrix.
    double
    get_adv_sev( Cnode*  cnode,
                 Thread* thrd );              ///< Returns the double representation of the value in the underlyoung AdvancedMatrix

    void
    set_adv_sev_adv( Cnode*  cnode,
                     Thread* thrd,
                     Value*  value );                                               ///< Saves the Value into AdvancedMatrix. It is the main "set_" method. All another call this method after needed transformations.
    Value*
    get_adv_sev_adv( Cnode*  cnode,
                     Thread* thrd );                                                ///< returns the Value from the AdvancedMatrix. It is the main "get_" method.


// -------- iterators for optimal data saving only along the calltree, threads can be saved in any sequance---------------
    vector<Cnode*>
    get_optimal_sequence( const vector<Cnode*>& );                                  ///< Returns the sequence of the cnodes. Using this sequence the neibourghood relation for calculation of excl/incl values is optimal.


    void
    setStrategy( CubeStrategy strategy ); ///< Setting the strategy for the metric

    void
    dropRow( Cnode* cnode );  ///< Attempts to drop the row to control the memory footprint manually. Only if strategy of metric allwos it

// ---------------------------------------------------------
// ----------- get incl/ecl severities ----------------------


    ///  Calculates inclusive value for given thread. Expands only calltree. Returns double representation of Value if possible.
    virtual double
    get_sev( Cnode*             cnode,
             CalculationFlavour cfv,
             Sysres*            sys,
             CalculationFlavour sf );


    /// Calculates inclusive value for given cnode and over all threads. Sums up over all threads. Assumed that all threads have continueos ids. Expands only calltree. Returns double representation of Value if possible.
    virtual double
    get_sev( Cnode*             cnode,
             CalculationFlavour cfv );

    ///  Calculates inclusive value for given thread. Expands only calltree.
    virtual Value*
    get_sev_adv( Cnode*             cnode,
                 CalculationFlavour cfv,
                 Sysres*            sys,
                 CalculationFlavour sf );


    /// Calculates inclusive value for given cnode and over all threads. Sums up over all threads. Assumed that all threads have continueos ids. Expands only calltree.
    virtual Value*
    get_sev_adv( Cnode*             cnode,
                 CalculationFlavour cfv );

    // -----------adv end -------------------------------

protected:
    std::string          disp_name;
    std::string          uniq_name;
    std::string          dtype;
    std::string          uom;
    std::string          val;      // optional
    std::string          url;
    std::string          descr;
    DataType             own_data_type;
    Value*               metric_value;           ///< Exemplar of the Value, which clones all others exemplats of Value for  calculation.
    FileFinder*          filefinder;
    std::string          expression;
    std::string          expression_init;
    CubePLMemoryManager* cubepl_memory_manager;
    GeneralEvaluation*   evaluation;
    GeneralEvaluation*   init_evaluation;
    bool                 initialized;
    bool                 active;
    TypeOfMetric         is_ghost;
    advMatrix*           adv_sev_mat;             ///< Matrix, saving all values.

    Cache* cache;                                 ///< Cache, performing internal caching.

    CubeStrategy my_strategy;                     ///< Saves metrics own strategy. Used for postponed initialization.

    unsigned ncid;                                ///< Layout of the metric
    unsigned ntid;                                ///< Layout of the metric


    map<unsigned, signed int> calltree_local_ids; ///< Used to optimise calculation of inclusive/exclusive values.
    map<unsigned, signed int> system_local_ids;   // at this time no local id are needed ... coz all threads are leaves in the tree...

    vector<Cnode*> cnodev;                        ///< Stores the flat list of cnodes
    vector<Cnode*> roots_cnodev;                  ///< Stores the list of roots of the call tree dimension

    vector<Machine*> roots_sysv;                  ///< Stores the roots gof system dimension
    vector<Thread*>  sysv;                        ///< Stores the flat list of threads.

    void
    printLocalCallTreeIDs();                        ///< Prints the metric own enumeration of calltree. Used for debugging.


    virtual void
    post_calculation_cleanup();

    virtual void
    pre_calculation_preparation( Cnode* );

    virtual void
    pre_calculation_preparation( Cnode*,
                                 Sysres* );


    void
    setup_cubepl_memory();

    virtual void
    setup_cubepl_memory_for_itself();

    virtual void
    setup_cubepl_memory_for_cnode( Cnode* );

    virtual void
    setup_cubepl_memory_for_region( Region* );

    virtual void
    setup_cubepl_memory_for_sysres( Sysres* );
};


/// Comparing of two metrics by comparing of their unique names.
inline bool
operator==( const Metric& a, const Metric& b )
{
    return a.get_uniq_name() == b.get_uniq_name();
}




typedef std::pair<Metric*, CalculationFlavour> metric_pair;             ///< Used for calculations of various values.
typedef std::vector<metric_pair>               list_of_metrics;         ///< Used to collect a list of metrics
}


#endif
