/****************************************************************************
**  CUBE        http://www.score-p.org/                                    **
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2014                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2014                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/

/**
 * \file
 * \brief Implementation of methods of the class SimpleCache
 */

#include "CubeSimpleCache.h"
#include "CubeCnode.h"
#include "CubeSysres.h"
#include "CubeValues.h"

using namespace std;
using namespace cube;


SimpleCache::~SimpleCache()
{
    // use Free call from Value to remove items "from memory" -> place them back to pool
    for ( cache_map::iterator iter = stn_container.begin(); iter != stn_container.end(); ++iter )
    {
        iter->second->Free();
    }
    for ( cache_map::iterator iter = sum_container.begin(); iter != sum_container.end(); ++iter )
    {
        iter->second->Free();
    }
}

Value*
SimpleCache::getCachedValue( Cnode* cnode, CalculationFlavour cf, Sysres* sysres, CalculationFlavour sf )
{
    simple_cache_key_t key = get_key( cnode, cf, sysres, sf );
    if ( key < 0 ) // do not cache ...
    {
        return 0;
    }

    Value* _value = 0;
    if ( sysres )
    {
        cache_map::iterator iter = stn_container.find( key );
        if ( iter != stn_container.end() )
        {
            _value = iter->second;
        }
    }
    else
    {
        cache_map::iterator iter = sum_container.find( key );
        if ( iter != sum_container.end() )
        {
            _value = iter->second;
        }
    }
    return _value ? _value->copy() : 0;     // pretend we do not
}

void
SimpleCache::setCachedValue( Value* value, Cnode* cnode, CalculationFlavour cf, Sysres* sysres, CalculationFlavour sf )
{
    simple_cache_key_t key = get_key( cnode, cf, sysres, sf );
    if ( key < 0 ) // do not cache ...
    {
        return;
    }

    // efficient "add if not yet in map" (S. Meyers, "Efficient STL", Item 24)
    if ( sysres )
    {
        cache_map::iterator iter = stn_container.lower_bound( key );
        if ( iter == stn_container.end() || stn_container.key_comp() ( key, iter->first ) )
        {
            stn_container.insert( iter, cache_map::value_type( key, value->copy() ) );
        }
    }
    else
    {
        cache_map::iterator iter = sum_container.lower_bound( key );
        if ( iter == sum_container.end() || sum_container.key_comp() ( key, iter->first ) )
        {
            sum_container.insert( iter, cache_map::value_type( key, value->copy() ) );
        }
    }
}

simple_cache_key_t
SimpleCache::get_key( Cnode* cnode, CalculationFlavour cf, Sysres* sysres, CalculationFlavour sf )
{
    if ( cf == myf )
    {
        return ( simple_cache_key_t )( -1 );
    }
    if ( !sysres )
    {
        return 2 * cnode->get_id() + cf;
    }
    if ( !sysres->isSystemTreeNode() )
    {
        return ( simple_cache_key_t )( -1 );
    }

    simple_cache_key_t Nc;
    if ( myf == CUBE_CALCULATE_EXCLUSIVE )
    {
        Nc = cnode->total_num_children();
    }
    else // myf == CUBE_CALCULATE_INCLUSIVE
    {
        Nc = cnode->num_children();
    }
    if ( Nc <= threshold )
    {
        return ( simple_cache_key_t )( -1 );
    }

    return 2 * number_loc * ( 2 * cnode->get_id() + cf ) +  2 * sysres->get_id() + sf;
}
